"""
Pipeline modules for the detection and interpolation of bad pixels.
"""

import warnings

from typing import Optional, Tuple

import numpy as np

from typeguard import typechecked

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.apply_func import bad_pixel_sigma_filter, image_interpolation, \
                                     replace_pixels, time_filter


class BadPixelSigmaFilterModule(ProcessingModule):
    """
    Pipeline module for finding bad pixels with a sigma filter and replacing them with the mean
    value of the surrounding pixels.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 map_out_tag: Optional[str] = None,
                 box: int = 9,
                 sigma: float = 5.,
                 iterate: int = 1) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output.
        map_out_tag : str, None
            Tag of the database entry with the bad pixel map that is written as output. No data
            is written if set to None. This output port can not be used if CPU > 1.
        box : int
            Size of the sigma filter. The area of the filter is equal to the squared value of
            *box*.
        sigma : float
            Sigma threshold.
        iterate : int
            Number of iterations.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        if map_out_tag is None:
            self.m_map_out_port = None
        else:
            self.m_map_out_port = self.add_output_port(map_out_tag)

        self.m_box = box
        self.m_sigma = sigma
        self.m_iterate = iterate

        if self.m_iterate < 1:
            raise ValueError('The argument of \'iterate\' should be 1 or larger.')

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Finds bad pixels with a sigma filter, replaces bad pixels with
        the mean value of the surrounding pixels, and writes the cleaned images to the database.

        Returns
        -------
        NoneType
            None
        """

        cpu = self._m_config_port.get_attribute('CPU')

        if cpu > 1 and self.m_map_out_port is not None:
            warnings.warn('The \'map_out_port\' can only be used if CPU = 1. No data will '
                          'be stored to this output port.')

            del self._m_output_ports[self.m_map_out_port.tag]
            self.m_map_out_port = None

        self.apply_function_to_images(bad_pixel_sigma_filter,
                                      self.m_image_in_port,
                                      self.m_image_out_port,
                                      'Bad pixel sigma filter',
                                      func_args=(self.m_box,
                                                 self.m_sigma,
                                                 self.m_iterate,
                                                 self.m_map_out_port))

        history = f'sigma = {self.m_sigma}'
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('BadPixelSigmaFilterModule', history)

        if self.m_map_out_port is not None:
            self.m_map_out_port.copy_attributes(self.m_image_in_port)
            self.m_map_out_port.add_history('BadPixelSigmaFilterModule', history)

        self.m_image_out_port.close_port()


class BadPixelMapModule(ProcessingModule):
    """
    Pipeline module to create a bad pixel map from the dark frames and flat fields.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 dark_in_tag: Optional[str],
                 flat_in_tag: Optional[str],
                 bp_map_out_tag: str,
                 dark_threshold: float = 0.2,
                 flat_threshold: float = 0.2) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        dark_in_tag : str, None
            Tag of the database entry with the dark frames that are read as input. Not read if set
            to None.
        flat_in_tag : str, None
            Tag of the database entry with the flat fields that are read as input. Not read if set
            to None.
        bp_map_out_tag : str
            Tag of the database entry with the bad pixel map that is written as output.
        dark_threshold : float
            Fractional threshold with respect to the maximum pixel value in the dark frame to flag
            bad pixels. Pixels `brighter` than the fractional threshold are flagged as bad.
        flat_threshold : float
            Fractional threshold with respect to the maximum pixel value in the flat field to flag
            bad pixels. Pixels `fainter` than the fractional threshold are flagged as bad.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        if dark_in_tag is None:
            self.m_dark_port = None
        else:
            self.m_dark_port = self.add_input_port(dark_in_tag)

        if flat_in_tag is None:
            self.m_flat_port = None
        else:
            self.m_flat_port = self.add_input_port(flat_in_tag)

        self.m_bp_map_out_port = self.add_output_port(bp_map_out_tag)

        self.m_dark_threshold = dark_threshold
        self.m_flat_threshold = flat_threshold

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Collapses a cube of dark frames and flat fields if needed, flags
        bad pixels by comparing the pixel values with the threshold times the maximum value, and
        writes a bad pixel map to the database. For the dark frame, pixel values larger than the
        threshold will be flagged while for the flat frame pixel values smaller than the threshold
        will be flagged.

        Returns
        -------
        NoneType
            None
        """

        if self.m_dark_port is not None:
            dark = self.m_dark_port.get_all()

            if dark.ndim == 3:
                dark = np.mean(dark, axis=0)

            max_dark = np.max(dark)

            print(f'Threshold dark frame = {max_dark*self.m_dark_threshold}')

            bpmap = np.ones(dark.shape)
            bpmap[np.where(dark > max_dark*self.m_dark_threshold)] = 0

        if self.m_flat_port is not None:
            flat = self.m_flat_port.get_all()

            if flat.ndim == 3:
                flat = np.mean(flat, axis=0)

            max_flat = np.max(flat)

            print(f'Threshold flat field (ADU) = {max_flat*self.m_flat_threshold:.2e}')

            if self.m_dark_port is None:
                bpmap = np.ones(flat.shape)

            bpmap[np.where(flat < max_flat*self.m_flat_threshold)] = 0

        if self.m_dark_port is not None and self.m_flat_port is not None:
            if not dark.shape == flat.shape:
                raise ValueError('Dark and flat images should have the same shape.')

        self.m_bp_map_out_port.set_all(bpmap, data_dim=3)

        if self.m_dark_port is not None:
            self.m_bp_map_out_port.copy_attributes(self.m_dark_port)
        elif self.m_flat_port is not None:
            self.m_bp_map_out_port.copy_attributes(self.m_flat_port)

        history = f'dark = {self.m_dark_threshold}, flat = {self.m_flat_threshold}'
        self.m_bp_map_out_port.add_history('BadPixelMapModule', history)

        self.m_bp_map_out_port.close_port()


class BadPixelInterpolationModule(ProcessingModule):
    """
    Pipeline module to interpolate bad pixels with spectral deconvolution.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 bad_pixel_map_tag: str,
                 image_out_tag: str,
                 iterations: int = 1000) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry with the images that are read as input.
        bad_pixel_map_tag : str
            Tag of the database entry with the bad pixel map that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output.
        iterations : int
            Number of iterations of the spectral deconvolution.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_bp_map_in_port = self.add_input_port(bad_pixel_map_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_iterations = iterations

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Interpolates bad pixels with an iterative spectral deconvolution.

        Returns
        -------
        NoneType
            None
        """

        bad_pixel_map = self.m_bp_map_in_port.get_all()[0, ]
        im_shape = self.m_image_in_port.get_shape()

        if self.m_iterations > im_shape[1]*im_shape[2]:
            raise ValueError('Maximum number of iterations needs to be smaller than the number of '
                             'pixels in the image.')

        if bad_pixel_map.shape[0] != im_shape[-2] or bad_pixel_map.shape[1] != im_shape[-1]:
            raise ValueError('The shape of the bad pixel map does not match the shape of the '
                             'images.')

        self.apply_function_to_images(image_interpolation,
                                      self.m_image_in_port,
                                      self.m_image_out_port,
                                      'Bad pixel interpolation',
                                      func_args=(self.m_iterations,
                                                 bad_pixel_map))

        history = f'iterations = {self.m_iterations}'
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('BadPixelInterpolationModule', history)
        self.m_image_out_port.close_port()


class BadPixelTimeFilterModule(ProcessingModule):
    """
    Pipeline module for finding bad pixels with a sigma filter along a pixel line in time. This
    module is suitable for removing bad pixels that are only present at a position in a small
    number of images, for example because a dither pattern has been applied. Pixel lines can be
    processed in parallel by setting the CPU keyword in the configuration file.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 sigma: Tuple[float, float] = (5., 5.)) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output.
        sigma : tuple(float, float)
            Lower and upper sigma threshold as (lower, upper).

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_sigma = sigma

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Finds bad pixels along a pixel line, replaces the bad pixels with
        the mean value of the pixels (excluding the bad pixels), and writes the cleaned images to
        the database.

        Returns
        -------
        NoneType
            None
        """

        print('Temporal filtering of bad pixels ...', end='')

        self.apply_function_in_time(time_filter,
                                    self.m_image_in_port,
                                    self.m_image_out_port,
                                    func_args=(self.m_sigma, ))

        print(' [DONE]')

        history = f'sigma = {self.m_sigma}'
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('BadPixelTimeFilterModule', history)
        self.m_image_out_port.close_port()


class ReplaceBadPixelsModule(ProcessingModule):
    """
    Pipeline module for replacing bad pixels with the mean are median value of the surrounding
    pixels. The bad pixels are selected from the input bad pixel map.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 map_in_tag: str,
                 image_out_tag: str,
                 size: int = 2,
                 replace: str = 'median') -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output.
        size : int
            Number of pixel lines around the bad pixel that are used to calculate the median or mean
            replacement value. For example, a 5x5 window is used if ``size=2``.
        replace : str
            Replace the bad pixel with the 'median', 'mean' or 'nan'.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_map_in_port = self.add_input_port(map_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_size = size
        self.m_replace = replace

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Masks the bad pixels with NaN and replaces the bad pixels with the
        mean or median value (excluding the bad pixels) within a window centered on the bad pixel.
        The original value is used if there are only NaNs within the window.

        Returns
        -------
        NoneType
            None
        """

        bpmap = self.m_map_in_port.get_all()[0, ]
        index = np.argwhere(bpmap == 0)

        self.apply_function_to_images(replace_pixels,
                                      self.m_image_in_port,
                                      self.m_image_out_port,
                                      'Running ReplaceBadPixelsModule',
                                      func_args=(index,
                                                 self.m_size,
                                                 self.m_replace))

        history = f'replace = {self.m_replace}'
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('ReplaceBadPixelsModule', history)
        self.m_image_out_port.close_port()



"""
Utilities for multiprocessing of stacks of images.
"""

import sys
import multiprocessing

from typing import Callable, List, Optional

import numpy as np

from typeguard import typechecked

from pynpoint.core.dataio import InputPort, OutputPort
from pynpoint.util.module import update_arguments
from pynpoint.util.multiproc import TaskInput, TaskResult, TaskCreator, TaskProcessor, \
                                    MultiprocessingCapsule, apply_function


class StackReader(TaskCreator):
    """
    Reader of task inputs for :class:`~pynpoint.util.multistack.StackProcessingCapsule`.
    Reads continuously stacks of images of a dataset and puts them into a task queue.
    """

    @typechecked
    def __init__(self,
                 data_port_in: InputPort,
                 tasks_queue_in: multiprocessing.JoinableQueue,
                 data_mutex_in: multiprocessing.Lock,
                 num_proc: int,
                 stack_size: int,
                 result_shape: tuple) -> None:
        """
        Parameters
        ----------
        data_port_in : pynpoint.core.dataio.InputPort
            Input port.
        tasks_queue_in : multiprocessing.queues.JoinableQueue
            Tasks queue.
        data_mutex_in : multiprocessing.synchronize.Lock
            A mutex shared with the writer to ensure that no read and write operations happen at
            the same time.
        num_proc : int
            Number of processors.
        stack_size: int
            Number of images per stack.
        result_shape : tuple(int, )
            Shape of the array with the output results (usually a stack of images).

        Returns
        -------
        NoneType
            None
        """

        super(StackReader, self).__init__(data_port_in, tasks_queue_in, data_mutex_in, num_proc)

        self.m_stack_size = stack_size
        self.m_result_shape = result_shape

    @typechecked
    def run(self) -> None:
        """
        Returns
        -------
        NoneType
            None
        """

        with self.m_data_mutex:
            self.m_data_in_port._check_status_and_activate()
            nimages = self.m_data_in_port.get_shape()[0]
            self.m_data_in_port.close_port()

        i = 0
        while i < nimages:
            j = min((i + self.m_stack_size), nimages)

            # lock mutex and read data
            with self.m_data_mutex:
                self.m_data_in_port._check_status_and_activate()
                tmp_data = self.m_data_in_port[i:j, ]  # read images from i to j
                self.m_data_in_port.close_port()

            # first dimension (start, stop, step)
            stack_slice = [(i, j, None)]

            # additional dimensions
            for _ in self.m_result_shape:
                stack_slice.append((None, None, None))

            param = (self.m_result_shape, tuple(stack_slice))
            self.m_task_queue.put(TaskInput(tmp_data, param))

            i = j

        self.create_poison_pills()


class StackTaskProcessor(TaskProcessor):
    """
    Processor of task inputs for :class:`~pynpoint.util.multistack.StackProcessingCapsule`. A
    processor applies a function on a stack of images.
    """

    @typechecked
    def __init__(self,
                 tasks_queue_in: multiprocessing.JoinableQueue,
                 result_queue_in: multiprocessing.JoinableQueue,
                 function: Callable,
                 function_args: Optional[tuple],
                 nimages: int) -> None:
        """
        Parameters
        ----------
        tasks_queue_in : multiprocessing.queues.JoinableQueue
            Tasks queue.
        result_queue_in : multiprocessing.queues.JoinableQueue
            Results queue.
        function : function
            Input function that is applied to the images.
        function_args : tuple, None
            Function arguments.
        nimages : int
            Total number of images.

        Returns
        -------
        NoneType
            None
        """

        super(StackTaskProcessor, self).__init__(tasks_queue_in, result_queue_in)

        self.m_function = function
        self.m_function_args = function_args
        self.m_nimages = nimages

    @typechecked
    def run_job(self,
                tmp_task: TaskInput) -> TaskResult:
        """
        Parameters
        ----------
        tmp_task : pynpoint.util.multiproc.TaskInput
            Task input with the subsets of images and the job parameters.

        Returns
        -------
        pynpoint.util.multiproc.TaskResult
            Task result.
        """

        result_nimages = tmp_task.m_input_data.shape[0]
        result_shape = tmp_task.m_job_parameter[0]

        # first dimension
        full_shape = [result_nimages]

        # additional dimensions
        for item in result_shape:
            full_shape.append(item)

        result_arr = np.zeros(full_shape)

        for i in range(result_nimages):
            # job parameter contains (result_shape, tuple(stack_slice))
            index = tmp_task.m_job_parameter[1][0][0] + i

            args = update_arguments(index, self.m_nimages, self.m_function_args)

            result_arr[i, ] = apply_function(tmp_task.m_input_data[i, ], i, self.m_function, args)

        sys.stdout.write('.')
        sys.stdout.flush()

        return TaskResult(result_arr, tmp_task.m_job_parameter[1])


class StackProcessingCapsule(MultiprocessingCapsule):
    """
    Capsule for parallel processing of stacks of images with the poison pill pattern. A function
    is applied in parallel to each stack of images.
    """

    @typechecked
    def __init__(self,
                 image_in_port: InputPort,
                 image_out_port: OutputPort,
                 num_proc: int,
                 function: Callable,
                 function_args: Optional[tuple],
                 stack_size: int,
                 result_shape: tuple,
                 nimages: int) -> None:
        """
        Parameters
        ----------
        image_in_port : pynpoint.core.dataio.InputPort
            Input port.
        image_out_port : pynpoint.core.dataio.OutputPort
            Output port.
        num_proc : int
            Number of processors.
        function : function
            Input function.
        function_args : tuple, None
            Function arguments.
        stack_size: int
            Number of images per stack.
        result_shape : tuple(int, )
            Shape of the array with output results (usually a stack of images).
        nimages : int
            Total number of images.

        Returns
        -------
        NoneType
            None
        """

        self.m_function = function
        self.m_function_args = function_args
        self.m_stack_size = stack_size
        self.m_result_shape = result_shape
        self.m_nimages = nimages

        super(StackProcessingCapsule, self).__init__(image_in_port, image_out_port, num_proc)

    @typechecked
    def create_processors(self) -> List[StackTaskProcessor]:
        """
        Returns
        -------
        list(pynpoint.util.multiproc.StackTaskProcessor, )
            List with instances of :class:`~pynpoint.util.multiproc.StackTaskProcessor`.
        """

        processors = []

        for _ in range(self.m_num_proc):

            processors.append(StackTaskProcessor(tasks_queue_in=self.m_tasks_queue,
                                                 result_queue_in=self.m_result_queue,
                                                 function=self.m_function,
                                                 function_args=self.m_function_args,
                                                 nimages=self.m_nimages))

        return processors

    @typechecked
    def init_creator(self,
                     image_in_port: InputPort) -> StackReader:
        """
        Parameters
        ----------
        image_in_port : pynpoint.core.dataio.InputPort
            Input port from where the subsets of images are read.

        Returns
        -------
        pynpoint.util.multistack.StackReader
            Reader of stacks of images.
        """

        return StackReader(data_port_in=image_in_port,
                           tasks_queue_in=self.m_tasks_queue,
                           data_mutex_in=self.m_data_mutex,
                           num_proc=self.m_num_proc,
                           stack_size=self.m_stack_size,
                           result_shape=self.m_result_shape)



import os

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.stacksubset import StackAndSubsetModule, StackCubesModule, \
                                            DerotateAndStackModule, CombineTagsModule
from pynpoint.util.tests import create_config, create_star_data, create_ifs_data, remove_test_data


class TestStackSubset:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_ifs_data(self.test_dir+'data_ifs')
        create_star_data(self.test_dir+'data')
        create_star_data(self.test_dir+'extra')

        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['data_ifs', 'extra', 'data'])

    def test_read_data(self) -> None:

        module = FitsReadingModule(name_in='read1',
                                   image_tag='images',
                                   input_dir=self.test_dir+'data',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read1')

        data = self.pipeline.get_data('images')
        assert np.mean(data) == pytest.approx(0.08722544528764692, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

        module = FitsReadingModule(name_in='read2',
                                   image_tag='extra',
                                   input_dir=self.test_dir+'extra',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read2')

        extra = self.pipeline.get_data('extra')
        assert data == pytest.approx(extra, rel=self.limit, abs=0.)

        module = FitsReadingModule(name_in='read_ifs',
                                   image_tag='images_ifs',
                                   input_dir=self.test_dir+'data_ifs',
                                   overwrite=True,
                                   check=True,
                                   ifs_data=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read_ifs')
        self.pipeline.set_attribute('images_ifs', 'PARANG', np.linspace(0., 180., 10), static=False)

        data = self.pipeline.get_data('images_ifs')
        assert np.sum(data) == pytest.approx(749.8396528807369, rel=self.limit, abs=0.)
        assert data.shape == (3, 10, 21, 21)

    def test_stack_and_subset(self) -> None:

        self.pipeline.set_attribute('images', 'PARANG', np.arange(10.), static=False)

        module = StackAndSubsetModule(name_in='stack1',
                                      image_in_tag='images',
                                      image_out_tag='stack1',
                                      random=4,
                                      stacking=2,
                                      combine='mean',
                                      max_rotation=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('stack1')

        data = self.pipeline.get_data('stack1')
        assert np.mean(data) == pytest.approx(0.08758276283743936, rel=self.limit, abs=0.)
        assert data.shape == (4, 11, 11)

        data = self.pipeline.get_data('header_stack1/INDEX')
        assert data == pytest.approx(np.arange(4), rel=self.limit, abs=0.)
        assert data.shape == (4, )

        data = self.pipeline.get_data('header_stack1/PARANG')
        assert data == pytest.approx([0.5, 2.5, 6.5, 8.5], rel=self.limit, abs=0.)
        assert data.shape == (4, )

    def test_stack_max_rotation(self) -> None:

        angles = np.arange(10.)
        angles[1:6] = 3.
        angles[9] = 50.

        self.pipeline.set_attribute('images', 'PARANG', angles, static=False)

        module = StackAndSubsetModule(name_in='stack2',
                                      image_in_tag='images',
                                      image_out_tag='stack2',
                                      random=None,
                                      stacking=2,
                                      combine='median',
                                      max_rotation=1.)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('stack2')

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'Testing of util.module.stack_angles has been ' \
                                             'limited, please use carefully.'

        data = self.pipeline.get_data('stack2')
        assert np.mean(data) == pytest.approx(0.08580759396987508, rel=self.limit, abs=0.)
        assert data.shape == (7, 11, 11)

        data = self.pipeline.get_data('header_stack2/INDEX')
        assert data == pytest.approx(np.arange(7), rel=self.limit, abs=0.)
        assert data.shape == (7, )

        data = self.pipeline.get_data('header_stack2/PARANG')
        assert data.shape == (7, )

        self.pipeline.set_attribute('images', 'PARANG', np.arange(10.), static=False)

    def test_stack_cube(self) -> None:

        module = StackCubesModule(name_in='stackcube',
                                  image_in_tag='images',
                                  image_out_tag='mean',
                                  combine='mean')

        self.pipeline.add_module(module)
        self.pipeline.run_module('stackcube')

        data = self.pipeline.get_data('mean')
        assert np.mean(data) == pytest.approx(0.08722544528764689, rel=self.limit, abs=0.)
        assert data.shape == (2, 11, 11)

        attribute = self.pipeline.get_attribute('mean', 'INDEX', static=False)
        assert np.mean(attribute) == pytest.approx(0.5, rel=self.limit, abs=0.)
        assert attribute.shape == (2, )

        attribute = self.pipeline.get_attribute('mean', 'NFRAMES', static=False)
        assert np.mean(attribute) == pytest.approx(1, rel=self.limit, abs=0.)
        assert attribute.shape == (2, )

    def test_derotate_and_stack(self) -> None:

        module = DerotateAndStackModule(name_in='derotate1',
                                        image_in_tag='images',
                                        image_out_tag='derotate1',
                                        derotate=True,
                                        stack='mean',
                                        extra_rot=10.)

        self.pipeline.add_module(module)
        self.pipeline.run_module('derotate1')

        data = self.pipeline.get_data('derotate1')
        assert np.mean(data) == pytest.approx(0.08709860116308817, rel=self.limit, abs=0.)
        assert data.shape == (1, 11, 11)

        module = DerotateAndStackModule(name_in='derotate2',
                                        image_in_tag='images',
                                        image_out_tag='derotate2',
                                        derotate=False,
                                        stack='median',
                                        extra_rot=0.)

        self.pipeline.add_module(module)
        self.pipeline.run_module('derotate2')

        data = self.pipeline.get_data('derotate2')
        assert np.mean(data) == pytest.approx(0.0861160094566323, rel=self.limit, abs=0.)
        assert data.shape == (1, 11, 11)

        data = self.pipeline.get_data('derotate2')
        assert np.mean(data) == pytest.approx(0.0861160094566323, rel=self.limit, abs=0.)
        assert data.shape == (1, 11, 11)

        module = DerotateAndStackModule(name_in='derotate_ifs1',
                                        image_in_tag='images_ifs',
                                        image_out_tag='derotate_ifs1',
                                        derotate=True,
                                        stack='mean',
                                        extra_rot=0.,
                                        dimension='time')

        self.pipeline.add_module(module)
        self.pipeline.run_module('derotate_ifs1')

        data = self.pipeline.get_data('derotate_ifs1')
        assert np.mean(data) == pytest.approx(0.1884438996655355, rel=self.limit, abs=0.)
        assert data.shape == (3, 1, 21, 21)

        module = DerotateAndStackModule(name_in='derotate_ifs2',
                                        image_in_tag='images_ifs',
                                        image_out_tag='derotate_ifs2',
                                        derotate=False,
                                        stack='median',
                                        extra_rot=0.,
                                        dimension='wavelength')

        self.pipeline.add_module(module)
        self.pipeline.run_module('derotate_ifs2')

        data = self.pipeline.get_data('derotate_ifs2')
        assert np.mean(data) == pytest.approx(0.055939644983170146, rel=self.limit, abs=0.)
        assert data.shape == (1, 10, 21, 21)

        module = DerotateAndStackModule(name_in='derotate_ifs3',
                                        image_in_tag='images_ifs',
                                        image_out_tag='derotate_ifs3',
                                        derotate=True,
                                        stack=None,
                                        extra_rot=0.,
                                        dimension='wavelength')

        self.pipeline.add_module(module)
        self.pipeline.run_module('derotate_ifs3')

        data = self.pipeline.get_data('derotate_ifs3')
        assert np.mean(data) == pytest.approx(0.05653316989966066, rel=self.limit, abs=0.)
        assert data.shape == (3, 10, 21, 21)

    def test_combine_tags(self) -> None:

        module = CombineTagsModule(image_in_tags=['images', 'extra'],
                                   check_attr=True,
                                   index_init=False,
                                   name_in='combine1',
                                   image_out_tag='combine1')

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('combine1')

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'The non-static keyword FILES is already used but ' \
                                             'with different values. It is advisable to only ' \
                                             'combine tags that descend from the same data set.'

        data = self.pipeline.get_data('combine1')
        assert np.mean(data) == pytest.approx(0.0872254452876469, rel=self.limit, abs=0.)
        assert data.shape == (20, 11, 11)

        data = self.pipeline.get_data('header_combine1/INDEX')
        assert data[19] == 9
        assert data.shape == (20, )

        module = CombineTagsModule(image_in_tags=['images', 'extra'],
                                   check_attr=False,
                                   index_init=True,
                                   name_in='combine2',
                                   image_out_tag='combine2')

        self.pipeline.add_module(module)
        self.pipeline.run_module('combine2')

        data = self.pipeline.get_data('combine1')
        extra = self.pipeline.get_data('combine2')
        assert data == pytest.approx(extra, rel=self.limit, abs=0.)

        data = self.pipeline.get_data('header_combine2/INDEX')
        assert data[19] == 19
        assert data.shape == (20, )



import os

import pytest
import h5py
import numpy as np

from astropy.io import fits

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.readwrite.fitswriting import FitsWritingModule
from pynpoint.processing.badpixel import BadPixelSigmaFilterModule
from pynpoint.util.tests import create_config, remove_test_data


class TestPypeline:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        np.random.seed(1)
        images = np.random.normal(loc=0, scale=2e-4, size=(5, 11, 11))

        hdu = fits.PrimaryHDU()
        header = hdu.header
        header['INSTRUME'] = 'IMAGER'
        header['HIERARCH ESO DET EXP NO'] = 1
        header['HIERARCH ESO DET NDIT'] = 5
        header['HIERARCH ESO INS PIXSCALE'] = 0.01
        header['HIERARCH ESO ADA POSANG'] = 10.
        header['HIERARCH ESO ADA POSANG END'] = 20.
        header['HIERARCH ESO SEQ CUMOFFSETX'] = 5.
        header['HIERARCH ESO SEQ CUMOFFSETY'] = 5.
        hdu.data = images
        hdu.writeto(self.test_dir+'images.fits')

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, files=['images.fits'])

    def test_create_default_config(self) -> None:

        with pytest.warns(UserWarning) as warning:
            Pypeline(self.test_dir, self.test_dir, self.test_dir)

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'Configuration file not found. Creating ' \
                                             'PynPoint_config.ini with default values ' \
                                             'in the working place.'

        with open(self.test_dir+'PynPoint_config.ini') as f_obj:
            count = 0
            for _ in f_obj:
                count += 1

        assert count == 25

    def test_create_none_config(self) -> None:

        file_obj = open(self.test_dir+'PynPoint_config.ini', 'w')
        file_obj.write('[header]\n\n')
        file_obj.write('INSTRUMENT: None\n')
        file_obj.write('NFRAMES: None\n')
        file_obj.write('EXP_NO: None\n')
        file_obj.write('NDIT: None\n')
        file_obj.write('PARANG_START: ESO ADA POSANG\n')
        file_obj.write('PARANG_END: None\n')
        file_obj.write('DITHER_X: None\n')
        file_obj.write('DITHER_Y: None\n')
        file_obj.write('DIT: None\n')
        file_obj.write('LATITUDE: None\n')
        file_obj.write('LONGITUDE: None\n')
        file_obj.write('PUPIL: None\n')
        file_obj.write('DATE: None\n')
        file_obj.write('RA: None\n')
        file_obj.write('DEC: None\n\n')
        file_obj.write('[settings]\n\n')
        file_obj.write('PIXSCALE: None\n')
        file_obj.write('MEMORY: None\n')
        file_obj.write('CPU: None\n')
        file_obj.close()

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        attribute = pipeline.get_attribute('config', 'MEMORY', static=True)
        assert attribute == 0

        attribute = pipeline.get_attribute('config', 'CPU', static=True)
        assert attribute == 0

        attribute = pipeline.get_attribute('config', 'PIXSCALE', static=True)
        assert attribute == pytest.approx(0., rel=self.limit, abs=0.)

        attribute = pipeline.get_attribute('config', 'INSTRUMENT', static=True)
        assert attribute == 'None'

        create_config(self.test_dir+'PynPoint_config.ini')

    def test_create_pipeline_path_missing(self) -> None:

        dir_non_exists = self.test_dir + 'none_dir/'
        dir_exists = self.test_dir

        with pytest.raises(AssertionError) as error:
            Pypeline(dir_non_exists, dir_exists, dir_exists)

        assert str(error.value) == 'The folder that was chosen for the working place does not ' \
                                   'exist: '+self.test_dir+'none_dir/.'

        with pytest.raises(AssertionError) as error:
            Pypeline(dir_exists, dir_non_exists, dir_exists)

        assert str(error.value) == 'The folder that was chosen for the input place does not ' \
                                   'exist: '+self.test_dir+'none_dir/.'

        with pytest.raises(AssertionError) as error:
            Pypeline(dir_exists, dir_exists, dir_non_exists)

        assert str(error.value) == 'The folder that was chosen for the output place does not ' \
                                   'exist: '+self.test_dir+'none_dir/.'

    def test_create_pipeline_existing_database(self) -> None:

        np.random.seed(1)
        images = np.random.normal(loc=0, scale=2e-4, size=(5, 11, 11))

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'w') as hdf_file:
            dset = hdf_file.create_dataset('images', data=images)
            dset.attrs['PIXSCALE'] = 0.01

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        data = pipeline.get_data('images')
        assert np.mean(data) == pytest.approx(1.1824138000882435e-05, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

        assert pipeline.get_attribute('images', 'PIXSCALE') == 0.01

        os.remove(self.test_dir+'PynPoint_database.hdf5')

    def test_create_pipeline_new_database(self) -> None:

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        pipeline.m_data_storage.open_connection()
        pipeline.m_data_storage.close_connection()

        del pipeline

        os.remove(self.test_dir+'PynPoint_database.hdf5')

    def test_add_module(self) -> None:

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        module = FitsReadingModule(name_in='read1',
                                   input_dir=None,
                                   image_tag='im_arr1')

        assert pipeline.add_module(module) is None

        module = FitsReadingModule(name_in='read2',
                                   input_dir=self.test_dir,
                                   image_tag='im_arr2')

        assert pipeline.add_module(module) is None

        with pytest.warns(UserWarning) as warning:
            pipeline.add_module(module)

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'Names of pipeline modules that are added to the ' \
                                             'Pypeline need to be unique. The current pipeline ' \
                                             'module, \'read2\', does already exist in the ' \
                                             'Pypeline dictionary so the previous module with ' \
                                             'the same name will be overwritten.'

        module = BadPixelSigmaFilterModule(name_in='badpixel',
                                           image_in_tag='im_arr1',
                                           image_out_tag='im_out')

        assert pipeline.add_module(module) is None

        module = FitsWritingModule(name_in='write1',
                                   file_name='result.fits',
                                   data_tag='im_arr1')

        assert pipeline.add_module(module) is None

        module = FitsWritingModule(name_in='write2',
                                   file_name='result.fits',
                                   data_tag='im_arr1',
                                   output_dir=self.test_dir)

        assert pipeline.add_module(module) is None

        assert pipeline.run() is None

        assert pipeline.get_module_names() == ['read1', 'read2', 'badpixel', 'write1', 'write2']

        os.remove(self.test_dir+'result.fits')
        os.remove(self.test_dir+'PynPoint_database.hdf5')

    def test_run_module(self) -> None:

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        module = FitsReadingModule(name_in='read',
                                   image_tag='im_arr')

        assert pipeline.add_module(module) is None
        assert pipeline.run_module('read') is None

        os.remove(self.test_dir+'PynPoint_database.hdf5')

    def test_run_module_wrong_tag(self) -> None:

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        module = FitsReadingModule(name_in='read')

        pipeline.add_module(module)

        module = FitsWritingModule(name_in='write',
                                   file_name='result.fits',
                                   data_tag='im_list')

        pipeline.add_module(module)

        module = BadPixelSigmaFilterModule(name_in='badpixel',
                                           image_in_tag='im_list',
                                           image_out_tag='im_out')

        pipeline.add_module(module)

        with pytest.raises(AttributeError) as error:
            pipeline.run_module('badpixel')

        assert str(error.value) == 'Pipeline module \'badpixel\' is looking for data under a ' \
                                   'tag which does not exist in the database.'

        with pytest.raises(AttributeError) as error:
            pipeline.run_module('write')

        assert str(error.value) == 'Pipeline module \'write\' is looking for data under a tag ' \
                                   'which does not exist in the database.'

        with pytest.raises(AttributeError) as error:
            pipeline.run()

        assert str(error.value) == 'Pipeline module \'write\' is looking for data under a tag ' \
                                   'which is not created by a previous module or the data does ' \
                                   'not exist in the database.'

        assert pipeline.validate_pipeline_module('test') == (False, 'test')

        os.remove(self.test_dir+'PynPoint_database.hdf5')

    def test_run_module_non_existing(self) -> None:

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        with pytest.warns(UserWarning) as warning:
            pipeline.run_module('test')

        assert len(warning) == 1
        assert warning[0].message.args[0] == 'Pipeline module \'test\' not found.'

        os.remove(self.test_dir+'PynPoint_database.hdf5')

    def test_remove_module(self) -> None:

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        module = FitsReadingModule(name_in='read')

        pipeline.add_module(module)

        module = BadPixelSigmaFilterModule(name_in='badpixel',
                                           image_in_tag='im_arr1',
                                           image_out_tag='im_out')

        pipeline.add_module(module)

        assert pipeline.get_module_names() == ['read', 'badpixel']
        assert pipeline.remove_module('read')

        assert pipeline.get_module_names() == ['badpixel']
        assert pipeline.remove_module('badpixel')

        with pytest.warns(UserWarning) as warning:
            pipeline.remove_module('test')

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'Pipeline module \'test\' is not found in the ' \
                                             'Pypeline dictionary so it could not be removed. ' \
                                             'The dictionary contains the following modules: [].' \

        os.remove(self.test_dir+'PynPoint_database.hdf5')

    def test_get_shape(self) -> None:

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        module = FitsReadingModule(name_in='read',
                                   image_tag='images')

        pipeline.add_module(module)
        pipeline.run_module('read')

        assert pipeline.get_shape('images') == (5, 11, 11)

    def test_get_tags(self) -> None:

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        assert pipeline.get_tags() == ['images']

    def test_list_attributes(self) -> None:

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        attr_dict = pipeline.list_attributes('images')

        assert len(attr_dict) == 11
        assert attr_dict['INSTRUMENT'] == 'IMAGER'
        assert attr_dict['PIXSCALE'] == 0.027
        assert attr_dict['NFRAMES'] == [5]
        assert attr_dict['PARANG_START'] == [10.]

    def test_set_and_get_attribute(self) -> None:

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        pipeline.set_attribute('images', 'PIXSCALE', 0.1, static=True)
        pipeline.set_attribute('images', 'PARANG', np.arange(1., 11., 1.), static=False)

        attribute = pipeline.get_attribute('images', 'PIXSCALE', static=True)
        assert attribute == pytest.approx(0.1, rel=self.limit, abs=0.)

        attribute = pipeline.get_attribute('images', 'PARANG', static=False)
        assert attribute == pytest.approx(np.arange(1., 11., 1.), rel=self.limit, abs=0.)

        pipeline.set_attribute('images', 'PARANG', np.arange(10., 21., 1.), static=False)

        attribute = pipeline.get_attribute('images', 'PARANG', static=False)
        assert attribute == pytest.approx(np.arange(10., 21., 1.), rel=self.limit, abs=0.)

    def test_get_data_range(self) -> None:

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        data = pipeline.get_data('images', data_range=(0, 2))

        assert data.shape == (2, 11, 11)

    def test_delete_data(self) -> None:

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        pipeline.delete_data('images')

        assert len(pipeline.get_tags()) == 0

    def test_delete_not_found(self) -> None:

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        with pytest.warns(UserWarning) as warning:
            pipeline.delete_data('images')

        assert len(warning) == 2

        assert warning[0].message.args[0] == 'Dataset \'images\' not found in the database.'
        assert warning[1].message.args[0] == 'Attributes of \'images\' not found in the database.'

    def test_omp_num_threads(self) -> None:

        os.environ['OMP_NUM_THREADS'] = '2'

        pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)
        pipeline.run()



"""
Pipeline modules to prepare the data for the PSF subtraction.
"""

import time
import warnings

from typing import Optional, Tuple

import numpy as np

from astropy.coordinates import EarthLocation
from astropy.time import Time
from typeguard import typechecked

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.module import progress, memory_frames
from pynpoint.util.image import create_mask, scale_image, shift_image


class PSFpreparationModule(ProcessingModule):
    """
    Module to prepare the data for PSF subtraction with PCA. The preparation steps include masking
    and an optional normalization.
    """

    __author__ = 'Markus Bonse, Tomas Stolker, Timothy Gebhard, Sven Kiefer'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 mask_out_tag: Optional[str] = None,
                 norm: bool = False,
                 resize: Optional[float] = None,
                 cent_size: Optional[float] = None,
                 edge_size: Optional[float] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry with images that is written as output.
        mask_out_tag : str, None, optional
            Tag of the database entry with the mask that is written as output. If set to None, no
            mask array is saved.
        norm : bool
            Normalize each image by its Frobenius norm. Only supported for 3D datasets (i.e.
            regular imaging).
        resize : float, None
            DEPRECATED. This parameter is currently ignored by the module and will be removed in a
            future version of PynPoint.
        cent_size : float, None, optional
            Radius of the central mask (in arcsec). No mask is used when set to None.
        edge_size : float, None, optional
            Outer radius (in arcsec) beyond which pixels are masked. No outer mask is used when set
            to None. If the value is larger than half the image size then it will be set to half
            the image size.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)

        if mask_out_tag is None:
            self.m_mask_out_port = None
        else:
            self.m_mask_out_port = self.add_output_port(mask_out_tag)

        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_cent_size = cent_size
        self.m_edge_size = edge_size
        self.m_norm = norm

        # Raise a DeprecationWarning if the resize argument is used
        if resize is not None:
            warnings.warn('The \'resize\' parameter has been deprecated. Its value is currently '
                          'being ignored, and the argument will be removed in a future version '
                          'of PynPoint.', DeprecationWarning)

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Masks and normalizes the images.
        Returns
        -------
        NoneType
            None
        """

        # Get the PIXSCALE and MEMORY attributes
        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')
        memory = self._m_config_port.get_attribute('MEMORY')

        # Get the numnber of dimensions and shape
        ndim = self.m_image_in_port.get_ndim()
        im_shape = self.m_image_in_port.get_shape()

        if ndim == 3:
            # Number of images
            nimages = im_shape[-3]

            # Split into batches to comply with memory constraints
            frames = memory_frames(memory, nimages)

        elif ndim == 4:
            # Process all wavelengths per exposure at once
            frames = np.linspace(0, im_shape[-3], im_shape[-3]+1)

        if self.m_norm and ndim == 4:
            warnings.warn('The \'norm\' parameter does not support 4D datasets and will therefore '
                          'be ignored.')

        # Convert m_cent_size and m_edge_size from arcseconds to pixels

        if self.m_cent_size is not None:
            self.m_cent_size /= pixscale

        if self.m_edge_size is not None:
            self.m_edge_size /= pixscale

        # Create 2D disk mask which will be applied to every frame
        mask = create_mask((int(im_shape[-2]), int(im_shape[-1])),
                           (self.m_cent_size, self.m_edge_size)).astype(bool)

        # Keep track of the normalization vectors in case we are normalizing the images (if
        # we are not normalizing, this list will remain empty)
        norms = list()

        start_time = time.time()

        # Run the PSFpreparationModule for each subset of frames
        for i in range(frames[:-1].size):
            # Print progress to command line
            progress(i, len(frames[:-1]), 'Preparing images for PSF subtraction...', start_time)

            if ndim == 3:
                # Get the images and ensure they have the correct 3D shape with the following
                # three dimensions: (batch_size, height, width)
                images = self.m_image_in_port[frames[i]:frames[i+1], ]

                if images.ndim == 2:
                    warnings.warn('The input data has 2 dimensions whereas 3 dimensions are '
                                  'required. An extra dimension has been added.')

                    images = images[np.newaxis, ...]

            elif ndim == 4:
                # Process all wavelengths per exposure at once
                images = self.m_image_in_port[:, i, ]

            # Apply the mask, i.e., set all pixels to 0 where the mask is False
            images[:, ~mask] = 0.

            # If desired, normalize the images using the Frobenius norm
            if self.m_norm and ndim == 3:
                im_norm = np.linalg.norm(images, ord='fro', axis=(1, 2))
                images /= im_norm[:, np.newaxis, np.newaxis]
                norms.append(im_norm)

            # Write processed images to output port
            if ndim == 3:
                self.m_image_out_port.append(images, data_dim=3)
            elif ndim == 4:
                self.m_image_out_port.append(images, data_dim=4)

        # Store information about mask
        if self.m_mask_out_port is not None:
            self.m_mask_out_port.set_all(mask)
            self.m_mask_out_port.copy_attributes(self.m_image_in_port)

        # Copy attributes from input port
        self.m_image_out_port.copy_attributes(self.m_image_in_port)

        # If the norms list is not empty (i.e., if we have computed the norm for every image),
        # we can also save the corresponding norm vector as an additional attribute
        if norms:
            self.m_image_out_port.add_attribute(name='norm',
                                                value=np.hstack(norms),
                                                static=False)

        # Save cent_size and edge_size as attributes to the output port
        if self.m_cent_size is not None:
            self.m_image_out_port.add_attribute(name='cent_size',
                                                value=self.m_cent_size * pixscale,
                                                static=True)

        if self.m_edge_size is not None:
            self.m_image_out_port.add_attribute(name='edge_size',
                                                value=self.m_edge_size * pixscale,
                                                static=True)


class AngleInterpolationModule(ProcessingModule):
    """
    Module for calculating the parallactic angle values by interpolating between the begin and end
    value of a data cube.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 data_tag: str) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        data_tag : str
            Tag of the database entry for which the parallactic angles are written as attributes.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_data_in_port = self.add_input_port(data_tag)
        self.m_data_out_port = self.add_output_port(data_tag)

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Calculates the parallactic angles of each frame by linearly
        interpolating between the start and end values of the data cubes. The values are written
        as attributes to *data_tag*. A correction of 360 deg is applied when the start and end
        values of the angles change sign at +/-180 deg.

        Returns
        -------
        NoneType
            None
        """

        parang_start = self.m_data_in_port.get_attribute('PARANG_START')
        parang_end = self.m_data_in_port.get_attribute('PARANG_END')

        steps = self.m_data_in_port.get_attribute('NFRAMES')

        if 'NDIT' in self.m_data_in_port.get_all_non_static_attributes():
            ndit = self.m_data_in_port.get_attribute('NDIT')

            if not np.all(ndit == steps):
                warnings.warn('There is a mismatch between the NDIT and NFRAMES values. The '
                              'parallactic angles are calculated with a linear interpolation by '
                              'using NFRAMES steps. A frame selection should be applied after '
                              'the parallactic angles are calculated.')

        new_angles = []

        start_time = time.time()
        for i, _ in enumerate(parang_start):
            progress(i, len(parang_start), 'Interpolating parallactic angles...', start_time)

            if parang_start[i] < -170. and parang_end[i] > 170.:
                parang_start[i] += 360.

            elif parang_end[i] < -170. and parang_start[i] > 170.:
                parang_end[i] += 360.

            if steps[i] == 1:
                new_angles = np.append(new_angles,
                                       [(parang_start[i] + parang_end[i])/2.])

            elif steps[i] != 1:
                new_angles = np.append(new_angles,
                                       np.linspace(parang_start[i],
                                                   parang_end[i],
                                                   num=steps[i]))

        self.m_data_out_port.add_attribute('PARANG',
                                           new_angles,
                                           static=False)


class SortParangModule(ProcessingModule):
    """
    Module to sort the images and attributes with increasing ``INDEX``.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Database tag with the input data.
        image_out_tag : str
            Database tag where the output data will be stored. Should be different from
            ``image_in_tag``.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Sorts the images and attributes with increasing ``INDEX``.
        Therefore, the images are sorted by there original (usually chronological) order.

        Returns
        -------
        NoneType
            None
        """

        memory = self._m_config_port.get_attribute('MEMORY')
        index = self.m_image_in_port.get_attribute('INDEX')

        ndim = self.m_image_in_port.get_ndim()
        nimages = self.m_image_in_port.get_shape()[-3]

        index_new = np.zeros(index.shape, dtype=int)

        if 'PARANG' in self.m_image_in_port.get_all_non_static_attributes():
            parang = self.m_image_in_port.get_attribute('PARANG')
            parang_new = np.zeros(parang.shape)

        else:
            parang_new = None

        if 'STAR_POSITION' in self.m_image_in_port.get_all_non_static_attributes():
            star = self.m_image_in_port.get_attribute('STAR_POSITION')
            star_new = np.zeros(star.shape)

        else:
            star_new = None

        index_sort = np.argsort(index)

        frames = memory_frames(memory, nimages)

        start_time = time.time()

        for i, _ in enumerate(frames[:-1]):
            progress(i, len(frames[:-1]), 'Sorting images in time...', start_time)

            index_new[frames[i]:frames[i+1]] = index[index_sort[frames[i]:frames[i+1]]]

            if parang_new is not None:
                parang_new[frames[i]:frames[i+1]] = parang[index_sort[frames[i]:frames[i+1]]]

            if star_new is not None:
                star_new[frames[i]:frames[i+1]] = star[index_sort[frames[i]:frames[i+1]]]

            # HDF5 indexing elements must be in increasing order
            for item in index_sort[frames[i]:frames[i+1]]:
                if ndim == 3:
                    self.m_image_out_port.append(self.m_image_in_port[item, ], data_dim=3)

                elif ndim == 4:
                    self.m_image_out_port.append(self.m_image_in_port[:, item, ], data_dim=4)

        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('SortParangModule', 'sorted by INDEX')
        self.m_image_out_port.add_attribute('INDEX', index_new, static=False)

        if parang_new is not None:
            self.m_image_out_port.add_attribute('PARANG', parang_new, static=False)

        if star_new is not None:
            self.m_image_out_port.add_attribute('STAR_POSITION', star_new, static=False)

        self.m_image_out_port.close_port()


class AngleCalculationModule(ProcessingModule):
    """
    Module for calculating the parallactic angles. The start time of the observation is taken and
    multiples of the exposure time are added to derive the parallactic angle of each frame inside
    the cube. Instrument specific overheads are included.
    """

    __author__ = 'Alexander Bohn, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 data_tag: str,
                 instrument: str = 'NACO') -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        data_tag : str
            Tag of the database entry for which the parallactic angles are written as attributes.
        instrument : str
            Instrument name ('NACO', 'SPHERE/IRDIS', or 'SPHERE/IFS').

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        # Parameters
        self.m_instrument = instrument

        # Set parameters according to choice of instrument
        if self.m_instrument == 'NACO':

            # pupil offset in degrees
            self.m_pupil_offset = 0.            # No offset here

            # no overheads in cube mode, since cube is read out after all individual exposures
            # see NACO manual page 62 (v102)
            self.m_O_START = 0.
            self.m_DIT_DELAY = 0.
            self.m_ROT = 0.

            # rotator offset in degrees
            self.m_rot_offset = 89.44           # According to NACO manual page 65 (v102)

        elif self.m_instrument == 'SPHERE/IRDIS':

            # pupil offset in degrees
            self.m_pupil_offset = -135.99       # According to SPHERE manual page 64 (v102)

            # overheads in cube mode (several NDITS) in hours
            self.m_O_START = 0.3 / 3600.        # According to SPHERE manual page 90/91 (v102)
            self.m_DIT_DELAY = 0.1 / 3600.      # According to SPHERE manual page 90/91 (v102)
            self.m_ROT = 0.838 / 3600.          # According to SPHERE manual page 90/91 (v102)

            # rotator offset in degrees
            self.m_rot_offset = 0.              # no offset here

        elif self.m_instrument == 'SPHERE/IFS':

            # pupil offset in degrees
            self.m_pupil_offset = -135.99 - 100.48  # According to SPHERE manual page 64 (v102)

            # overheads in cube mode (several NDITS) in hours
            self.m_O_START = 0.3 / 3600.            # According to SPHERE manual page 90/91 (v102)
            self.m_DIT_DELAY = 0.2 / 3600.          # According to SPHERE manual page 90/91 (v102)
            self.m_ROT = 1.65 / 3600.               # According to SPHERE manual page 90/91 (v102)

            # rotator offset in degrees
            self.m_rot_offset = 0.                  # no offset here

        else:
            raise ValueError('The instrument argument should be set to either \'NACO\', '
                             '\'SPHERE/IRDIS\', or \'SPHERE/IFS\'.')

        self.m_data_in_port = self.add_input_port(data_tag)
        self.m_data_out_port = self.add_output_port(data_tag)

    @typechecked
    def _attribute_check(self,
                         ndit: np.ndarray,
                         steps: np.ndarray) -> None:

        if not np.all(ndit == steps):
            warnings.warn('There is a mismatch between the NDIT and NFRAMES values. A frame '
                          'selection should be applied after the parallactic angles are '
                          'calculated.')

        if self.m_instrument == 'SPHERE/IFS':
            warnings.warn('AngleCalculationModule has not been tested for SPHERE/IFS data.')

        if self.m_instrument in ('SPHERE/IRDIS', 'SPHERE/IFS'):

            if self._m_config_port.get_attribute('RA') != 'ESO INS4 DROT2 RA':

                warnings.warn('For SPHERE data it is recommended to use the header keyword '
                              '\'ESO INS4 DROT2 RA\' to specify the object\'s right ascension. '
                              'The input will be parsed accordingly. Using the regular '
                              '\'RA\' keyword will lead to wrong parallactic angles.')

            if self._m_config_port.get_attribute('DEC') != 'ESO INS4 DROT2 DEC':

                warnings.warn('For SPHERE data it is recommended to use the header keyword '
                              '\'ESO INS4 DROT2 DEC\' to specify the object\'s declination. '
                              'The input will be parsed accordingly. Using the regular '
                              '\'DEC\' keyword will lead to wrong parallactic angles.')

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Calculates the parallactic angles from the position of the object
        on the sky and the telescope location on earth. The start of the observation is used to
        extrapolate for the observation time of each individual image of a data cube. The values
        are written as PARANG attributes to *data_tag*.

        Returns
        -------
        NoneType
            None
        """

        # Load cube sizes
        steps = self.m_data_in_port.get_attribute('NFRAMES')
        ndit = self.m_data_in_port.get_attribute('NDIT')

        self._attribute_check(ndit, steps)

        # Load exposure time [hours]
        exptime = self.m_data_in_port.get_attribute('DIT')/3600.

        # Load telescope location
        tel_lat = self.m_data_in_port.get_attribute('LATITUDE')
        tel_lon = self.m_data_in_port.get_attribute('LONGITUDE')

        # Load temporary target position
        tmp_ra = self.m_data_in_port.get_attribute('RA')
        tmp_dec = self.m_data_in_port.get_attribute('DEC')

        # Parse to degree depending on instrument
        if 'SPHERE' in self.m_instrument:

            # get sign of declination
            tmp_dec_sign = np.sign(tmp_dec)
            tmp_dec = np.abs(tmp_dec)

            # parse RA
            tmp_ra_s = tmp_ra % 100
            tmp_ra_m = ((tmp_ra - tmp_ra_s) / 1e2) % 100
            tmp_ra_h = ((tmp_ra - tmp_ra_s - tmp_ra_m * 1e2) / 1e4)

            # parse DEC
            tmp_dec_s = tmp_dec % 100
            tmp_dec_m = ((tmp_dec - tmp_dec_s) / 1e2) % 100
            tmp_dec_d = ((tmp_dec - tmp_dec_s - tmp_dec_m * 1e2) / 1e4)

            # get RA and DEC in degree
            ra = (tmp_ra_h + tmp_ra_m / 60. + tmp_ra_s / 3600.) * 15.
            dec = tmp_dec_sign * (tmp_dec_d + tmp_dec_m / 60. + tmp_dec_s / 3600.)

        else:
            ra = tmp_ra
            dec = tmp_dec

        # Load start times of exposures
        obs_dates = self.m_data_in_port.get_attribute('DATE')

        # Load pupil positions during observations
        if self.m_instrument == 'NACO':
            pupil_pos = self.m_data_in_port.get_attribute('PUPIL')

        elif self.m_instrument == 'SPHERE/IRDIS':
            pupil_pos = np.zeros(steps.shape)

        elif self.m_instrument == 'SPHERE/IFS':
            pupil_pos = np.zeros(steps.shape)

        new_angles = np.array([])
        pupil_pos_arr = np.array([])

        start_time = time.time()

        # Calculate parallactic angles for each cube
        for i, tmp_steps in enumerate(steps):
            progress(i, len(steps), 'Calculating parallactic angles...', start_time)

            t = Time(obs_dates[i].decode('utf-8'),
                     location=EarthLocation(lat=tel_lat, lon=tel_lon))

            sid_time = t.sidereal_time('apparent').value

            # Extrapolate sideral times from start time of the cube for each frame of it
            sid_time_arr = np.linspace(sid_time+self.m_O_START,
                                       (sid_time+self.m_O_START) +
                                       (exptime+self.m_DIT_DELAY + self.m_ROT)*(tmp_steps-1),
                                       tmp_steps)

            # Convert to degrees
            sid_time_arr_deg = sid_time_arr * 15.

            # Calculate hour angle in degrees
            hour_angle = sid_time_arr_deg - ra[i]

            # Conversion to radians:
            hour_angle_rad = np.deg2rad(hour_angle)
            dec_rad = np.deg2rad(dec[i])
            lat_rad = np.deg2rad(tel_lat)

            p_angle = np.arctan2(np.sin(hour_angle_rad),
                                 (np.cos(dec_rad)*np.tan(lat_rad) -
                                  np.sin(dec_rad)*np.cos(hour_angle_rad)))

            new_angles = np.append(new_angles, np.rad2deg(p_angle))
            pupil_pos_arr = np.append(pupil_pos_arr, np.ones(tmp_steps)*pupil_pos[i])

        # Correct for rotator (SPHERE) or pupil offset (NACO)
        if self.m_instrument == 'NACO':
            # See NACO manual page 65 (v102)
            new_angles_corr = new_angles - (90. + (self.m_rot_offset-pupil_pos_arr))

        elif self.m_instrument == 'SPHERE/IRDIS':
            # See SPHERE manual page 64 (v102)
            new_angles_corr = new_angles - self.m_pupil_offset

        elif self.m_instrument == 'SPHERE/IFS':
            # See SPHERE manual page 64 (v102)
            new_angles_corr = new_angles - self.m_pupil_offset

        indices = np.where(new_angles_corr < -180.)[0]
        if indices.size > 0:
            new_angles_corr[indices] += 360.

        indices = np.where(new_angles_corr > 180.)[0]
        if indices.size > 0:
            new_angles_corr[indices] -= 360.

        self.m_data_out_port.add_attribute('PARANG', new_angles_corr, static=False)


class SDIpreparationModule(ProcessingModule):
    """
    Module for preparing continuum frames for dual-band simultaneous differential imaging.
    """

    __author__ = 'Gabriele Cugno, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 wavelength: Tuple[float, float],
                 width: Tuple[float, float]) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output. Should be different from
            *image_in_tag*.
        wavelength : tuple(float, float)
            The central wavelengths of the line and continuum filter, (line, continuum), in
            arbitrary but identical units.
        width : tuple(float, float)
            The equivalent widths of the line and continuum filter, (line, continuum), in
            arbitrary but identical units.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_line_wvl = wavelength[0]
        self.m_cnt_wvl = wavelength[1]

        self.m_line_width = width[0]
        self.m_cnt_width = width[1]

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Normalizes the images for the different filter widths,
        upscales the images, and crops the images to the initial image shape in order to
        align the PSF patterns.

        Returns
        -------
        NoneType
            None
        """

        wvl_factor = self.m_line_wvl/self.m_cnt_wvl
        width_factor = self.m_line_width/self.m_cnt_width

        nimages = self.m_image_in_port.get_shape()[0]

        start_time = time.time()
        for i in range(nimages):
            progress(i, nimages, 'Preparing images for dual-band SDI...', start_time)

            image = self.m_image_in_port[i, ]

            im_scale = width_factor * scale_image(image, wvl_factor, wvl_factor)

            if i == 0:
                npix_del = im_scale.shape[-1] - image.shape[-1]

                if npix_del % 2 == 0:
                    npix_del_a = int(npix_del/2)
                    npix_del_b = int(npix_del/2)

                else:
                    npix_del_a = int((npix_del-1)/2)
                    npix_del_b = int((npix_del+1)/2)

            im_crop = im_scale[npix_del_a:-npix_del_b, npix_del_a:-npix_del_b]

            if npix_del % 2 == 1:
                im_crop = shift_image(im_crop, (-0.5, -0.5), interpolation='spline')

            self.m_image_out_port.append(im_crop, data_dim=3)

        history = f'(line, continuum) = ({self.m_line_wvl}, {self.m_cnt_wvl})'
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('SDIpreparationModule', history)
        self.m_image_in_port.close_port()



"""
Functions for stellar extraction.
"""

import math
import time

from typing import Optional, Tuple, Union

import cv2
import numpy as np

from typeguard import typechecked

from pynpoint.core.dataio import InputPort
from pynpoint.util.image import crop_image, center_pixel
from pynpoint.util.module import progress


@typechecked
def locate_star(image: np.ndarray,
                center: Optional[tuple],
                width: Optional[int],
                fwhm: Optional[int]) -> np.ndarray:
    """
    Function to locate the star by finding the brightest pixel.

    Parameters
    ----------
    image : numpy.ndarray
        Input image (2D).
    center : tuple(int, int), None
        Pixel center (y, x) of the subframe. The full image is used if set to None.
    width : int, None
        The width (pix) of the subframe. The full image is used if set to None.
    fwhm : int, None
        Full width at half maximum (pix) of the Gaussian kernel. Not used if set to None.

    Returns
    -------
    numpy.ndarray
        Position (y, x) of the brightest pixel.
    """

    if width is not None:
        if center is None:
            center = center_pixel(image)

        image = crop_image(image, center, width)

    if fwhm is None:
        smooth = np.copy(image)

    else:
        sigma = fwhm / math.sqrt(8. * math.log(2.))
        kernel = (fwhm * 2 + 1, fwhm * 2 + 1)
        smooth = cv2.GaussianBlur(image, kernel, sigma)

    # argmax[0] is the y position and argmax[1] is the y position
    argmax = np.asarray(np.unravel_index(smooth.argmax(), smooth.shape))

    if center is not None and width is not None:
        argmax[0] += center[0] - (image.shape[0] - 1) // 2  # y
        argmax[1] += center[1] - (image.shape[1] - 1) // 2  # x

    return argmax


@typechecked
def star_positions(input_port: InputPort,
                   fwhm: Optional[int],
                   position: Optional[Union[Tuple[int, int, float],
                                            Tuple[None, None, float],
                                            Tuple[int, int, None]]] = None) -> np.ndarray:
    """
    Function to return the position of the star in each image.

    Parameters
    ----------
    input_port : pynpoint.core.dataio.InputPort
        Input port where the images are stored.
    fwhm : int, None
        The FWHM (pix) of the Gaussian kernel that is used to smooth the images before the
        brightest pixel is located. No smoothing is applied if set to None.
    position : tuple(int, int, int), None
        Subframe that is selected to search for the star. The tuple contains the center (pix)
        and size (pix) (pos_x, pos_y, size). Setting `position` to None will use the full
        image to search for the star. If `position=(None, None, size)` then the center of the
        image will be used. If `position=(pos_x, pos_y, None)` then a fixed position is used
        for the aperture.

    Returns
    -------
    numpy.ndarray
        Positions (y, x) of the brightest pixel.
    """

    nimages = input_port.get_shape()[0]
    starpos = np.zeros((nimages, 2), dtype=np.int64)

    if position is not None and position[2] is None:
        # [y. x] position
        starpos[:, 0] = position[1]
        starpos[:, 1] = position[0]

    else:
        center = None
        width = None

        if position is not None:
            width = position[2]

            if position[0] is not None and position[1] is not None:
                center = position[0:2]

        start_time = time.time()

        for i in range(nimages):
            progress(i, nimages, 'Locating stellar position...', start_time)

            # [y. x] position
            starpos[i, :] = locate_star(input_port[i, ], center, width, fwhm)

    return starpos



"""
Pipeline modules for PSF subtraction.
"""

import time
import math
import warnings

from copy import deepcopy
from typing import List, Optional, Tuple, Union

import numpy as np

from scipy.ndimage import rotate
from sklearn.decomposition import PCA
from typeguard import typechecked

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.apply_func import subtract_psf
from pynpoint.util.module import progress
from pynpoint.util.multipca import PcaMultiprocessingCapsule
from pynpoint.util.residuals import combine_residuals
from pynpoint.util.postproc import postprocessor
from pynpoint.util.sdi import scaling_factors


class PcaPsfSubtractionModule(ProcessingModule):
    """
    Pipeline module for PSF subtraction with principal component analysis (PCA). The module can
    be used for ADI, RDI (see ``subtract_mean`` parameter), SDI, and ASDI. The residuals are
    calculated in parallel for the selected numbers of principal components. This may require
    a large amount of memory in case the stack of input images is very large. The number of
    processes can therefore be set with the ``CPU`` keyword in the configuration file.
    """

    __author__ = 'Markus Bonse, Tomas Stolker, Sven Kiefer'

    @typechecked
    def __init__(self,
                 name_in: str,
                 images_in_tag: str,
                 reference_in_tag: str,
                 res_mean_tag: Optional[str] = None,
                 res_median_tag: Optional[str] = None,
                 res_weighted_tag: Optional[str] = None,
                 res_rot_mean_clip_tag: Optional[str] = None,
                 res_arr_out_tag: Optional[str] = None,
                 basis_out_tag: Optional[str] = None,
                 pca_numbers: Union[range,
                                    List[int],
                                    np.ndarray,
                                    Tuple[range, range],
                                    Tuple[List[int], List[int]],
                                    Tuple[np.ndarray, np.ndarray]] = range(1, 21),
                 extra_rot: float = 0.,
                 subtract_mean: bool = True,
                 processing_type: str = 'ADI') -> None:
        """
        Parameters
        ----------
        name_in : str
            Name tag of the pipeline module.
        images_in_tag : str
            Database entry with the images from which the PSF model will be subtracted.
        reference_in_tag : str
            Database entry with the reference images from which the PSF model is created. Usually
            ``reference_in_tag`` is the same as ``images_in_tag``, but a different dataset can be
            used as reference images in case of RDI.
        res_mean_tag : str, None
            Database entry where the the mean-collapsed residuals will be stored. The residuals are
            not calculated and stored if set to None.
        res_median_tag : str, None
            Database entry where the the median-collapsed residuals will be stored. The residuals
            are not calculated and stored if set to None.
        res_weighted_tag : str, None
            Database entry where the the noise-weighted residuals will be stored (see Bottom et al.
            2017). The residuals are not calculated and stored if set to None.
        res_rot_mean_clip_tag : str, None
            Tag of the database entry of the clipped mean residuals. Not calculated if set to
            None.
        res_arr_out_tag : str, None
            Database entry where the derotated, but not collapsed, residuals are stored. The number
            of principal components is was used is appended to the ``res_arr_out_tag``. The
            residuals are not stored if set to None. This parameter is not supported with
            multiprocessing (i.e. ``CPU`` > 1). For IFS data and if the processing type is either
            ADI+SDI or SDI+ADI the residuals can only be calculated if exactly 1 principal component
            for each ADI and SDI is given with the pca_numbers parameter.
        basis_out_tag : str, None
            Database entry where the principal components are stored. The data is not stored if set
            to None. Only supported for imaging data with ``processing_type='ADI'``.
        pca_numbers : range, list(int), np.ndarray, tuple(range, range), tuple[list(int),
                      list(int)), tuple(np.ndarray, np.ndarray))
            Number of principal components that are used for the PSF model. With ADI or SDI, a
            single list/range/array needs to be provided while for SDI+ADI or ADI+SDI a tuple is
            required with twice a list/range/array.
        extra_rot : float
            Additional rotation angle of the images (deg).
        subtract_mean : bool
            The mean of the science and reference images is subtracted from the corresponding
            stack, before the PCA basis is constructed and fitted. Set the argument to ``False``
            for RDI, that is, in case ``reference_in_tag`` is different from ``images_in_tag``
            and there is no or limited field rotation. The parameter is only supported with
            ``processing_type='ADI'``.
        processing_type : str
            Post-processing type:
                - ADI: Angular differential imaging. Can be used both on imaging and IFS datasets.
                  This argument is also used for RDI, in which case the ``PARANG`` attribute should
                  contain zeros a derotation angles (e.g. with
                  :func:`~pynpoint.core.pypeline.Pypeline.set_attribute` or
                  :class:`~pynpoint.readwrite.attr_writing.ParangWritingModule`). The collapsed
                  residuals are stored as 3D dataset with one image per principal component.
                - SDI: Spectral differential imaging. Can only be applied on IFS datasets. The
                  collapsed residuals are stored as $D dataset with one image per wavelength and
                  principal component.
                - SDI+ADI: Spectral and angular differential imaging. Can only be applied on IFS
                  datasets. The collapsed residuals are stored as 5D datasets with one image per
                  wavelength and each of the principal components.
                - ADI+SDI: Angular and spectral differential imaging. Can only be applied on IFS
                  datasets. The collapsed residuals are stored as 5D datasets with one image per
                  wavelength and each of the principal components.
                - CODI: Perform ADI and SDI simultaniously. Usually requires a lot more PC than
                  the other techniques. The collapsed residuals are stored as 4D datasets with one
                  image per wavelength and each of the principal components.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_pca_numbers = pca_numbers

        if isinstance(pca_numbers, tuple):
            self.m_components = (np.sort(np.atleast_1d(pca_numbers[0])),
                                 np.sort(np.atleast_1d(pca_numbers[1])))

        else:
            self.m_components = np.sort(np.atleast_1d(pca_numbers))
            self.m_pca = PCA(n_components=np.amax(self.m_components), svd_solver='arpack')

        self.m_extra_rot = extra_rot
        self.m_subtract_mean = subtract_mean
        self.m_processing_type = processing_type

        self.m_reference_in_port = self.add_input_port(reference_in_tag)
        self.m_star_in_port = self.add_input_port(images_in_tag)

        if res_mean_tag is None:
            self.m_res_mean_out_port = None
        else:
            self.m_res_mean_out_port = self.add_output_port(res_mean_tag)

        if res_median_tag is None:
            self.m_res_median_out_port = None
        else:
            self.m_res_median_out_port = self.add_output_port(res_median_tag)

        if res_weighted_tag is None:
            self.m_res_weighted_out_port = None
        else:
            self.m_res_weighted_out_port = self.add_output_port(res_weighted_tag)

        if res_rot_mean_clip_tag is None:
            self.m_res_rot_mean_clip_out_port = None
        else:
            self.m_res_rot_mean_clip_out_port = self.add_output_port(res_rot_mean_clip_tag)

        if res_arr_out_tag is None:
            self.m_res_arr_out_ports = None
        else:
            if isinstance(self.m_components, tuple):
                self.m_res_arr_out_ports = self.add_output_port(res_arr_out_tag)
            else:
                self.m_res_arr_out_ports = {}

                for pca_number in self.m_components:
                    self.m_res_arr_out_ports[pca_number] = self.add_output_port(
                        res_arr_out_tag + str(pca_number))

        if basis_out_tag is None:
            self.m_basis_out_port = None
        else:
            self.m_basis_out_port = self.add_output_port(basis_out_tag)

        if self.m_processing_type in ['ADI', 'SDI', 'CODI']:
            if not isinstance(self.m_components, (range, list, np.ndarray)):
                raise ValueError(f'The post-processing type \'{self.m_processing_type}\' requires '
                                 f'a single range/list/array as argument for \'pca_numbers\'.')

        elif self.m_processing_type in ['SDI+ADI', 'ADI+SDI']:
            if not isinstance(self.m_components, tuple):
                raise ValueError(f'The post-processing type \'{self.m_processing_type}\' requires '
                                 f'a tuple for with twice a range/list/array as argument for '
                                 f'\'pca_numbers\'.')

            if res_arr_out_tag is not None and len(self.m_components[0]) + \
                    len(self.m_components[1]) != 2:
                raise ValueError(f'If the post-processing type \'{self.m_processing_type}\' '
                                 'is selected, residuals can only be calculated if no more than '
                                 '1 principal component for ADI and SDI is given.')
        else:
            raise ValueError('Please select a valid post-processing type.')

    @typechecked
    def _run_multi_processing(self,
                              star_reshape: np.ndarray,
                              im_shape: tuple,
                              indices: Optional[np.ndarray]) -> None:
        """
        Internal function to create the residuals, derotate the images, and write the output
        using multiprocessing.
        """

        cpu = self._m_config_port.get_attribute('CPU')
        parang = -1.*self.m_star_in_port.get_attribute('PARANG') + self.m_extra_rot

        if self.m_ifs_data:
            if 'WAVELENGTH' in self.m_star_in_port.get_all_non_static_attributes():
                wavelength = self.m_star_in_port.get_attribute('WAVELENGTH')

            else:
                raise ValueError('The wavelengths are not found. These should be stored '
                                 'as the \'WAVELENGTH\' attribute.')

            scales = scaling_factors(wavelength)

        else:
            scales = None

        if self.m_processing_type in ['ADI', 'SDI', 'CODI']:
            pca_first = self.m_components
            pca_secon = [-1]  # Not used

        elif self.m_processing_type in ['SDI+ADI', 'ADI+SDI']:
            pca_first = self.m_components[0]
            pca_secon = self.m_components[1]

        if self.m_ifs_data:
            if self.m_processing_type in ['ADI', 'SDI', 'CODI']:
                res_shape = (len(pca_first), len(wavelength), im_shape[-2], im_shape[-1])

            elif self.m_processing_type in ['SDI+ADI', 'ADI+SDI']:
                res_shape = (len(pca_first), len(pca_secon), len(wavelength),
                             im_shape[-2], im_shape[-1])

        else:
            res_shape = (len(self.m_components), im_shape[1], im_shape[2])

        tmp_output = np.zeros(res_shape)

        if self.m_res_mean_out_port is not None:
            self.m_res_mean_out_port.set_all(tmp_output, keep_attributes=False)

        if self.m_res_median_out_port is not None:
            self.m_res_median_out_port.set_all(tmp_output, keep_attributes=False)

        if self.m_res_weighted_out_port is not None:
            self.m_res_weighted_out_port.set_all(tmp_output, keep_attributes=False)

        if self.m_res_rot_mean_clip_out_port is not None:
            self.m_res_rot_mean_clip_out_port.set_all(tmp_output, keep_attributes=False)

        self.m_star_in_port.close_port()
        self.m_reference_in_port.close_port()

        if self.m_res_mean_out_port is not None:
            self.m_res_mean_out_port.close_port()

        if self.m_res_median_out_port is not None:
            self.m_res_median_out_port.close_port()

        if self.m_res_weighted_out_port is not None:
            self.m_res_weighted_out_port.close_port()

        if self.m_res_rot_mean_clip_out_port is not None:
            self.m_res_rot_mean_clip_out_port.close_port()

        if self.m_basis_out_port is not None:
            self.m_basis_out_port.close_port()

        capsule = PcaMultiprocessingCapsule(self.m_res_mean_out_port,
                                            self.m_res_median_out_port,
                                            self.m_res_weighted_out_port,
                                            self.m_res_rot_mean_clip_out_port,
                                            cpu,
                                            deepcopy(self.m_components),
                                            deepcopy(self.m_pca),
                                            deepcopy(star_reshape),
                                            deepcopy(parang),
                                            deepcopy(scales),
                                            im_shape,
                                            indices,
                                            self.m_processing_type)

        capsule.run()

    @typechecked
    def _run_single_processing(self,
                               star_reshape: np.ndarray,
                               im_shape: tuple,
                               indices: Optional[np.ndarray]) -> None:
        """
        Internal function to create the residuals, derotate the images, and write the output
        using a single process.
        """

        start_time = time.time()

        # Get the parallactic angles
        parang = -1.*self.m_star_in_port.get_attribute('PARANG') + self.m_extra_rot

        if self.m_ifs_data:
            # Get the wavelengths
            if 'WAVELENGTH' in self.m_star_in_port.get_all_non_static_attributes():
                wavelength = self.m_star_in_port.get_attribute('WAVELENGTH')

            else:
                raise ValueError('The wavelengths are not found. These should be stored '
                                 'as the \'WAVELENGTH\' attribute.')

            # Calculate the wavelength ratios
            scales = scaling_factors(wavelength)

        else:
            scales = None

        if self.m_processing_type in ['ADI', 'SDI', 'CODI']:
            pca_first = self.m_components
            pca_secon = [-1]  # Not used
            n_pca_total = len(pca_first)

        elif self.m_processing_type in ['SDI+ADI', 'ADI+SDI']:
            pca_first = self.m_components[0]
            pca_secon = self.m_components[1]
            n_pca_total = len(pca_first) + len(pca_secon)

        # Setup output arrays

        out_array_res = np.zeros(im_shape)

        if self.m_ifs_data:
            if self.m_processing_type in ['ADI', 'SDI', 'CODI']:
                res_shape = (len(pca_first), len(wavelength), im_shape[-2], im_shape[-1])

            elif self.m_processing_type in ['SDI+ADI', 'ADI+SDI']:
                res_shape = (len(pca_first), len(pca_secon), len(wavelength),
                             im_shape[-2], im_shape[-1])

        else:
            res_shape = (len(pca_first), im_shape[-2], im_shape[-1])

        out_array_mean = np.zeros(res_shape)
        out_array_medi = np.zeros(res_shape)
        out_array_weig = np.zeros(res_shape)
        out_array_clip = np.zeros(res_shape)

        # loop over all different combination of pca_numbers and applying the reductions
        for i, pca_1 in enumerate(pca_first):
            for j, pca_2 in enumerate(pca_secon):
                progress(i+j, n_pca_total, 'Creating residuals...', start_time)

                # process images
                residuals, res_rot = postprocessor(images=star_reshape,
                                                   angles=parang,
                                                   scales=scales,
                                                   pca_number=(pca_1, pca_2),
                                                   pca_sklearn=self.m_pca,
                                                   im_shape=im_shape,
                                                   indices=indices,
                                                   processing_type=self.m_processing_type)

                # 1.) derotated residuals
                if self.m_res_arr_out_ports is not None:
                    if not self.m_ifs_data:
                        self.m_res_arr_out_ports[pca_1].set_all(res_rot)
                        self.m_res_arr_out_ports[pca_1].copy_attributes(self.m_star_in_port)
                        self.m_res_arr_out_ports[pca_1].add_history(
                            'PcaPsfSubtractionModule', f'max PC number = {pca_first}')

                    else:
                        out_array_res = residuals

                # 2.) mean residuals
                if self.m_res_mean_out_port is not None:
                    if self.m_processing_type in ['SDI+ADI', 'ADI+SDI']:
                        out_array_mean[i, j] = combine_residuals(method='mean',
                                                                 res_rot=res_rot,
                                                                 angles=parang)

                    else:
                        out_array_mean[i] = combine_residuals(method='mean',
                                                              res_rot=res_rot,
                                                              angles=parang)

                # 3.) median residuals
                if self.m_res_median_out_port is not None:
                    if self.m_processing_type in ['SDI+ADI', 'ADI+SDI']:
                        out_array_medi[i, j] = combine_residuals(method='median',
                                                                 res_rot=res_rot,
                                                                 angles=parang)

                    else:
                        out_array_medi[i] = combine_residuals(method='median',
                                                              res_rot=res_rot,
                                                              angles=parang)

                # 4.) noise-weighted residuals
                if self.m_res_weighted_out_port is not None:
                    if self.m_processing_type in ['SDI+ADI', 'ADI+SDI']:
                        out_array_weig[i, j] = combine_residuals(method='weighted',
                                                                 res_rot=res_rot,
                                                                 residuals=residuals,
                                                                 angles=parang)

                    else:
                        out_array_weig[i] = combine_residuals(method='weighted',
                                                              res_rot=res_rot,
                                                              residuals=residuals,
                                                              angles=parang)

                # 5.) clipped mean residuals
                if self.m_res_rot_mean_clip_out_port is not None:
                    if self.m_processing_type in ['SDI+ADI', 'ADI+SDI']:
                        out_array_clip[i, j] = combine_residuals(method='clipped',
                                                                 res_rot=res_rot,
                                                                 angles=parang)

                    else:
                        out_array_clip[i] = combine_residuals(method='clipped',
                                                              res_rot=res_rot,
                                                              angles=parang)

        # Configurate data output according to the processing type
        # 1.) derotated residuals
        if self.m_res_arr_out_ports is not None and self.m_ifs_data:
            if pca_secon[0] == -1:
                history = f'max PC number = {pca_first}'

            else:
                history = f'max PC number = {pca_first} / {pca_secon}'

            # squeeze out_array_res to reduce dimensionallity as the residuals of
            # SDI+ADI and ADI+SDI are always of the form (1, 1, ...)
            squeezed = np.squeeze(out_array_res)

            if isinstance(self.m_components, tuple):
                self.m_res_arr_out_ports.set_all(squeezed, data_dim=squeezed.ndim)
                self.m_res_arr_out_ports.copy_attributes(self.m_star_in_port)
                self.m_res_arr_out_ports.add_history('PcaPsfSubtractionModule', history)

            else:
                for i, pca in enumerate(self.m_components):
                    self.m_res_arr_out_ports[pca].append(squeezed[i])
                    self.m_res_arr_out_ports[pca].add_history('PcaPsfSubtractionModule', history)

        # 2.) mean residuals
        if self.m_res_mean_out_port is not None:
            self.m_res_mean_out_port.set_all(out_array_mean,
                                             data_dim=out_array_mean.ndim)

        # 3.) median residuals
        if self.m_res_median_out_port is not None:
            self.m_res_median_out_port.set_all(out_array_medi,
                                               data_dim=out_array_medi.ndim)

        # 4.) noise-weighted residuals
        if self.m_res_weighted_out_port is not None:
            self.m_res_weighted_out_port.set_all(out_array_weig,
                                                 data_dim=out_array_weig.ndim)

        # 5.) clipped mean residuals
        if self.m_res_rot_mean_clip_out_port is not None:
            self.m_res_rot_mean_clip_out_port.set_all(out_array_clip,
                                                      data_dim=out_array_clip.ndim)

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Subtracts the mean of the image stack from all images, reshapes
        the stack of images into a 2D array, uses singular value decomposition to construct the
        orthogonal basis set, calculates the PCA coefficients for each image, subtracts the PSF
        model, and writes the residuals as output.

        Returns
        -------
        NoneType
            None
        """

        print('Input parameters:')
        print(f'   - Post-processing type: {self.m_processing_type}')
        print(f'   - Number of principal components: {self.m_pca_numbers}')
        print(f'   - Subtract mean: {self.m_subtract_mean}')
        print(f'   - Extra rotation (deg): {self.m_extra_rot}')

        cpu = self._m_config_port.get_attribute('CPU')

        if cpu > 1 and self.m_res_arr_out_ports is not None:
            warnings.warn('Multiprocessing not possible if \'res_arr_out_tag\' is not set '
                          'to None.')

        # Read the data
        star_data = self.m_star_in_port.get_all()
        im_shape = star_data.shape

        # Parse input processing types to internal processing types
        if star_data.ndim == 3:
            self.m_ifs_data = False

        elif star_data.ndim == 4:
            self.m_ifs_data = True

        else:
            raise ValueError(f'The input data has {star_data.ndim} dimensions while only 3 or 4 '
                             f' are supported by the pipeline module.')

        if self.m_processing_type == 'ADI' and not self.m_ifs_data:
            # select the first image and get the unmasked image indices
            im_star = star_data[0, ].reshape(-1)
            indices = np.where(im_star != 0.)[0]

            # reshape the star data and select the unmasked pixels
            star_reshape = star_data.reshape(im_shape[0], im_shape[1]*im_shape[2])
            star_reshape = star_reshape[:, indices]

            if self.m_reference_in_port.tag == self.m_star_in_port.tag:
                ref_reshape = deepcopy(star_reshape)

            else:
                ref_data = self.m_reference_in_port.get_all()
                ref_shape = ref_data.shape

                if ref_shape[-2:] != im_shape[-2:]:
                    raise ValueError('The image size of the science data and the reference data '
                                     'should be identical.')

                # reshape reference data and select the unmasked pixels
                ref_reshape = ref_data.reshape(ref_shape[0], ref_shape[1]*ref_shape[2])
                ref_reshape = ref_reshape[:, indices]

            # subtract mean from science data, if required
            if self.m_subtract_mean:
                mean_star = np.mean(star_reshape, axis=0)
                star_reshape -= mean_star

            # subtract mean from reference data
            mean_ref = np.mean(ref_reshape, axis=0)
            ref_reshape -= mean_ref

            # create the PCA basis
            print('Constructing PSF model...', end='')
            self.m_pca.fit(ref_reshape)

            # add mean of reference array as 1st PC and orthogonalize it with respect to
            # the other principal components
            if not self.m_subtract_mean:
                mean_ref_reshape = mean_ref.reshape((1, mean_ref.shape[0]))

                q_ortho, _ = np.linalg.qr(np.vstack((mean_ref_reshape,
                                                     self.m_pca.components_[:-1, ])).T)

                self.m_pca.components_ = q_ortho.T

            print(' [DONE]')

            if self.m_basis_out_port is not None:
                pc_size = self.m_pca.components_.shape[0]

                basis = np.zeros((pc_size, im_shape[1]*im_shape[2]))
                basis[:, indices] = self.m_pca.components_
                basis = basis.reshape((pc_size, im_shape[1], im_shape[2]))

                self.m_basis_out_port.set_all(basis)

        else:
            # This setup is used for SDI processes. No preparations are possible because SDI/ADI
            # combinations are case specific and need to be conducted in pca_psf_subtraction.
            self.m_pca = None
            indices = None
            star_reshape = star_data

        # Running a single processing PCA analysis
        if cpu == 1 or self.m_res_arr_out_ports is not None:
            self._run_single_processing(star_reshape, im_shape, indices)

        # Running multiprocessed PCA analysis
        else:
            print('Creating residuals', end='')
            self._run_multi_processing(star_reshape, im_shape, indices)
            print(' [DONE]')

        # write history
        if isinstance(self.m_components, tuple):
            history = f'max PC number = {np.amax(self.m_components[0])} / ' \
                      f'{np.amax(self.m_components[1])}'

        else:
            history = f'max PC number = {np.amax(self.m_components)}'

        # save history for all other ports
        if self.m_res_mean_out_port is not None:
            self.m_res_mean_out_port.copy_attributes(self.m_star_in_port)
            self.m_res_mean_out_port.add_history('PcaPsfSubtractionModule', history)

        if self.m_res_median_out_port is not None:
            self.m_res_median_out_port.copy_attributes(self.m_star_in_port)
            self.m_res_median_out_port.add_history('PcaPsfSubtractionModule', history)

        if self.m_res_weighted_out_port is not None:
            self.m_res_weighted_out_port.copy_attributes(self.m_star_in_port)
            self.m_res_weighted_out_port.add_history('PcaPsfSubtractionModule', history)

        if self.m_res_rot_mean_clip_out_port is not None:
            self.m_res_rot_mean_clip_out_port.copy_attributes(self.m_star_in_port)
            self.m_res_rot_mean_clip_out_port.add_history('PcaPsfSubtractionModule', history)

        self.m_star_in_port.close_port()


class ClassicalADIModule(ProcessingModule):
    """
    Pipeline module for PSF subtraction with classical ADI by subtracting a median-combined
    reference image. A rotation threshold can be set for a fixed separation to prevent
    self-subtraction.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 res_out_tag: str,
                 stack_out_tag: str,
                 threshold: Optional[Tuple[float, float, float]],
                 nreference: Optional[int] = None,
                 residuals: str = 'median',
                 extra_rot: float = 0.) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry with the science images that are read as input.
        res_out_tag : str
            Tag of the database entry with the residuals of the PSF subtraction that are written
            as output.
        stack_out_tag : str
            Tag of the database entry with the stacked residuals that are written as output.
        threshold : tuple(float, float, float), None
            Tuple with the separation for which the angle threshold is optimized (arcsec), FWHM of
            the PSF (arcsec), and the threshold (FWHM) for the selection of the reference images.
            No threshold is used if set to None.
        nreference : int, None
            Number of reference images, closest in line to the science image. All images are used if
            *threshold* is None or *nreference* is None.
        residuals : str
            Method used for combining the residuals ('mean', 'median', 'weighted', or 'clipped').
        extra_rot : float
            Additional rotation angle of the images (deg).

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_res_out_port = self.add_output_port(res_out_tag)
        self.m_stack_out_port = self.add_output_port(stack_out_tag)

        self.m_threshold = threshold
        self.m_nreference = nreference
        self.m_extra_rot = extra_rot
        self.m_residuals = residuals

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Selects for each image the reference images closest in line while
        taking into account a rotation threshold for a fixed separation, median-combines the
        references images, and subtracts the reference image from each image separately.
        Alternatively, a single, median-combined reference image can be created and subtracted from
        all images. All images are used if the rotation condition can not be met. Both the
        individual residuals (before derotation) and the stacked residuals are stored.

        Returns
        -------
        NoneType
            None
        """

        parang = -1.*self.m_image_in_port.get_attribute('PARANG') + self.m_extra_rot

        nimages = self.m_image_in_port.get_shape()[0]

        if self.m_threshold:
            parang_thres = 2.*math.atan2(self.m_threshold[2]*self.m_threshold[1],
                                         2.*self.m_threshold[0])
            parang_thres = math.degrees(parang_thres)
            reference = None

        else:
            parang_thres = None
            reference = self.m_image_in_port.get_all()
            reference = np.median(reference, axis=0)

        ang_diff = np.zeros((nimages, parang.shape[0]))

        for i in range(nimages):
            ang_diff[i, :] = np.abs(parang[i] - parang)

        self.apply_function_to_images(subtract_psf,
                                      self.m_image_in_port,
                                      self.m_res_out_port,
                                      'Classical ADI',
                                      func_args=(parang_thres,
                                                 self.m_nreference,
                                                 reference,
                                                 ang_diff,
                                                 self.m_image_in_port))

        self.m_res_in_port = self.add_input_port(self.m_res_out_port._m_tag)
        im_res = self.m_res_in_port.get_all()

        res_rot = np.zeros(im_res.shape)
        for i, item in enumerate(parang):
            res_rot[i, ] = rotate(im_res[i, ], item, reshape=False)

        stack = combine_residuals(self.m_residuals,
                                  res_rot,
                                  residuals=im_res,
                                  angles=parang)

        self.m_stack_out_port.set_all(stack)

        if self.m_threshold:
            history = f'threshold [deg] = {parang_thres:.2f}'
        else:
            history = 'threshold [deg] = None'

        self.m_res_out_port.copy_attributes(self.m_image_in_port)
        self.m_res_out_port.add_history('ClassicalADIModule', history)

        self.m_stack_out_port.copy_attributes(self.m_image_in_port)
        self.m_stack_out_port.add_history('ClassicalADIModule', history)

        self.m_res_out_port.close_port()



"""
Pipeline modules for resizing of images.
"""

import math
import time

from typing import Tuple, Union

import numpy as np

from typeguard import typechecked

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.apply_func import image_scaling
from pynpoint.util.image import crop_image
from pynpoint.util.module import memory_frames, progress


class CropImagesModule(ProcessingModule):
    """
    Pipeline module for cropping of images around a given position.
    """

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 size: float,
                 center: Union[Tuple[int, int], None]) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output. Should be different from
            *image_in_tag*.
        size : float
            New image size (arcsec). The same size will be used for both image dimensions.
        center : tuple(int, int), None
            Tuple (x0, y0) with the new image center. Python indexing starts at 0. The center of
            the input images will be used when *center* is set to *None*. Note that if the image
            is even-sized, it is not possible to a uniquely define a pixel position in the center
            of the image. The image center is determined (with pixel precision) with the
            :func:`~pynpoint.util.image.center_pixel` function.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_size = size
        self.m_center = center

        if self.m_center is not None:
            self.m_center = (self.m_center[1], self.m_center[0])  # (y, x)

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Decreases the image size by cropping around an given position.
        The module always returns odd-sized images.

        Returns
        -------
        NoneType
            None
        """

        # Get memory and number of images to split the frames into chunks
        memory = self._m_config_port.get_attribute('MEMORY')
        nimages = self.m_image_in_port.get_shape()[0]

        # Get the numnber of dimensions and shape
        ndim = self.m_image_in_port.get_ndim()
        im_shape = self.m_image_in_port.get_shape()

        if ndim == 3:
            # Number of images
            nimages = im_shape[-3]

            # Split into batches to comply with memory constraints
            frames = memory_frames(memory, nimages)

        elif ndim == 4:
            # Process all wavelengths per exposure at once
            frames = np.linspace(0, im_shape[-3], im_shape[-3]+1)

        # Convert size parameter from arcseconds to pixels
        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')
        print(f'New image size (arcsec) = {self.m_size}')
        self.m_size = int(math.ceil(self.m_size / pixscale))
        print(f'New image size (pixels) = {self.m_size}')

        if self.m_center is not None:
            print(f'New image center (x, y) = ({self.m_center[1]}, {self.m_center[0]})')

        # Crop images chunk by chunk
        start_time = time.time()
        for i in range(len(frames[:-1])):

            # Update progress bar
            progress(i, len(frames[:-1]), 'Cropping images...', start_time)

            # Select images in the current chunk
            if ndim == 3:
                images = self.m_image_in_port[frames[i]:frames[i+1], ]

            elif ndim == 4:
                # Process all wavelengths per exposure at once
                images = self.m_image_in_port[:, i, ]

            # crop images according to input parameters
            images = crop_image(images, self.m_center, self.m_size, copy=False)

            # Write processed images to output port
            if ndim == 3:
                self.m_image_out_port.append(images, data_dim=3)
            elif ndim == 4:
                self.m_image_out_port.append(images, data_dim=4)

        # Save history and copy attributes
        history = f'image size (pix) = {self.m_size}'
        self.m_image_out_port.add_history('CropImagesModule', history)
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.close_port()


class ScaleImagesModule(ProcessingModule):
    """
    Pipeline module for rescaling of an image.
    """

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 scaling: Union[Tuple[float, float, float],
                                Tuple[None, None, float],
                                Tuple[float, float, None]],
                 pixscale: bool = True) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output. Should be different from
            ``image_in_tag``.
        scaling : tuple(float, float, float)
            Tuple with the scaling factors for the image size and flux, (scaling_x, scaling_y,
            scaling_flux). Upsampling and downsampling of the image corresponds to
            ``scaling_x/y`` > 1 and 0 < ``scaling_x/y`` < 1, respectively.
        pixscale : bool
            Adjust the pixel scale by the average scaling in x and y direction.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        if scaling[0] is None:
            self.m_scaling_x = 1.
        else:
            self.m_scaling_x = scaling[0]

        if scaling[1] is None:
            self.m_scaling_y = 1.
        else:
            self.m_scaling_y = scaling[1]

        if scaling[2] is None:
            self.m_scaling_flux = 1.
        else:
            self.m_scaling_flux = scaling[2]

        self.m_pixscale = pixscale

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Rescales an image with a fifth order spline interpolation and a
        reflecting boundary condition.

        Returns
        -------
        NoneType
            None
        """

        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')

        self.apply_function_to_images(image_scaling,
                                      self.m_image_in_port,
                                      self.m_image_out_port,
                                      'Scaling images',
                                      func_args=(self.m_scaling_y,
                                                 self.m_scaling_x,
                                                 self.m_scaling_flux))

        history = f'scaling = ({self.m_scaling_x:.2f}, {self.m_scaling_y:.2f}, ' \
                  f'{self.m_scaling_flux:.2f})'

        self.m_image_out_port.add_history('ScaleImagesModule', history)
        self.m_image_out_port.copy_attributes(self.m_image_in_port)

        if self.m_pixscale:
            mean_scaling = (self.m_scaling_x+self.m_scaling_y)/2.
            self.m_image_out_port.add_attribute('PIXSCALE', pixscale/mean_scaling)

        self.m_image_out_port.close_port()


class AddLinesModule(ProcessingModule):
    """
    Module to add lines of pixels to increase the size of an image.
    """

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 lines: Tuple[int, int, int, int]) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output, including the images with
            increased size. Should be different from *image_in_tag*.
        lines : tuple(int, int, int, int)
            The number of lines that are added in left, right, bottom, and top direction.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_lines = np.asarray(lines)

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Adds lines of zero-value pixels to increase the size of an image.

        Returns
        -------
        NoneType
            None
        """

        memory = self._m_config_port.get_attribute('MEMORY')
        nimages = self.m_image_in_port.get_shape()[0]
        frames = memory_frames(memory, nimages)

        shape_in = self.m_image_in_port.get_shape()

        shape_out = (shape_in[-2]+int(self.m_lines[2])+int(self.m_lines[3]),
                     shape_in[-1]+int(self.m_lines[0])+int(self.m_lines[1]))

        self.m_lines[1] = shape_out[1] - self.m_lines[1]  # right side of image
        self.m_lines[3] = shape_out[0] - self.m_lines[3]  # top side of image

        start_time = time.time()

        for i in range(len(frames[:-1])):
            progress(i, len(frames[:-1]), 'Adding lines...', start_time)

            image_in = self.m_image_in_port[frames[i]:frames[i+1], ]

            image_out = np.zeros((frames[i+1]-frames[i], shape_out[0], shape_out[1]))

            image_out[:,
                      int(self.m_lines[2]):int(self.m_lines[3]),
                      int(self.m_lines[0]):int(self.m_lines[1])] = image_in

            self.m_image_out_port.append(image_out, data_dim=3)

        history = f'number of lines = {self.m_lines}'
        self.m_image_out_port.add_history('AddLinesModule', history)
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.close_port()


class RemoveLinesModule(ProcessingModule):
    """
    Module to decrease the dimensions of an image by removing lines of pixels.
    """

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 lines: Tuple[int, int, int, int]) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output, including the images with
            decreased size. Should be different from *image_in_tag*.
        lines : tuple(int, int, int, int)
            The number of lines that are removed in left, right, bottom, and top direction.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_lines = lines

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Removes the lines given by *lines* from each frame.

        Returns
        -------
        NoneType
            None
        """

        memory = self._m_config_port.get_attribute('MEMORY')
        nimages = self.m_image_in_port.get_shape()[0]
        frames = memory_frames(memory, nimages)

        start_time = time.time()

        for i in range(len(frames[:-1])):
            progress(i, len(frames[:-1]), 'Removing lines...', start_time)

            image_in = self.m_image_in_port[frames[i]:frames[i+1], ]

            image_out = image_in[:,
                                 int(self.m_lines[2]):image_in.shape[1]-int(self.m_lines[3]),
                                 int(self.m_lines[0]):image_in.shape[2]-int(self.m_lines[1])]

            self.m_image_out_port.append(image_out, data_dim=3)

        history = f'number of lines = {self.m_lines}'
        self.m_image_out_port.add_history('RemoveLinesModule', history)
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.close_port()



"""
Module for reading FITS files.
"""

import os
import time
import warnings

from typing import List, Optional, Tuple, Union

import numpy as np

from astropy.io import fits
from typeguard import typechecked

from pynpoint.core.processing import ReadingModule
from pynpoint.util.attributes import set_static_attr, set_nonstatic_attr, set_extra_attr
from pynpoint.util.module import progress


class FitsReadingModule(ReadingModule):
    """
    Reads FITS files from the given *input_dir* or the default directory of the Pypeline. The FITS
    files need to contain either single images (2D) or cubes of images (3D). Individual images
    should have the same shape and type. The header of the FITS is scanned for the required static
    attributes (should be identical for each FITS file) and non-static attributes. Static entries
    will be saved as HDF5 attributes while non-static attributes will be saved as separate data
    sets in a subfolder of the database named *header_* + image_tag. If the FITS files in the input
    directory have changing static attributes or the shape of the input images is changing a
    warning appears. FitsReadingModule overwrites by default all existing data with the same tags
    in the central database.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 input_dir: Optional[str] = None,
                 image_tag: str = 'im_arr',
                 overwrite: bool = True,
                 check: bool = True,
                 filenames: Optional[Union[str, List[str]]] = None,
                 ifs_data: bool = False) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        input_dir : str, None
            Input directory where the FITS files are located. If not specified the Pypeline default
            directory is used.
        image_tag : str
            Tag of the read data in the HDF5 database. Non static header information is stored with
            the tag: *header_* + image_tag / header_entry_name.
        overwrite : bool
            Overwrite existing data and header in the central database.
        check : bool
            Print a warning if certain attributes from the configuration file are not present in
            the FITS header. If set to `False`, attributes are still written to the dataset but
            there will be no warning if a keyword is not found in the FITS header.
        filenames : str, list(str, ), None
            If a string, then a path of a text file should be provided. This text file should
            contain a list of FITS files. If a list, then the paths of the FITS files should be
            provided directly. If set to None, the FITS files in the `input_dir` are read. All
            paths should be provided either relative to the Python working folder (i.e., the folder
            where Python is executed) or as absolute paths.
        ifs_data : bool
            Import IFS data which is stored as a 4D array with the wavelength and temporal
            dimensions as first and second dimension, respectively. If set to ``False`` (default),
            the data is imported as a 3D array with the temporal dimension as first dimension.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in, input_dir=input_dir)

        self.m_image_out_port = self.add_output_port(image_tag)

        self.m_overwrite = overwrite
        self.m_check = check
        self.m_filenames = filenames
        self.m_ifs_data = ifs_data

    @typechecked
    def read_single_file(self,
                         fits_file: str,
                         overwrite_tags: list) -> Tuple[fits.header.Header, tuple]:
        """
        Function which reads a single FITS file and appends it to the database. The function gets
        a list of *overwriting_tags*. If a new key (header entry or image data) is found that is
        not on this list the old entry is overwritten if *self.m_overwrite* is active. After
        replacing the old entry the key is added to the *overwriting_tags*. This procedure
        guaranties that all previous database information, that does not belong to the new data
        set that is read by FitsReadingModule is replaced and the rest is kept.

        Parameters
        ----------
        fits_file : str
            Absolute path and filename of the FITS file.
        overwrite_tags : list(str, )
            The list of database tags that will not be overwritten.

        Returns
        -------
        astropy.io.fits.header.Header
            FITS header.
        tuple(int, )
            Image shape.
        """

        hdu_list = fits.open(fits_file)

        if hdu_list[0].data is not None:
            images = hdu_list[0].data.byteswap().newbyteorder()

        elif len(hdu_list) > 1:
            for i, item in enumerate(hdu_list[1:]):
                if isinstance(item, fits.hdu.image.ImageHDU):
                    warnings.simplefilter('always', UserWarning)

                    warnings.warn(f"No data was found in the PrimaryHDU "
                                  f"so reading data from the ImageHDU "
                                  f"at number {i+1} instead.")

                    images = hdu_list[i+1].data.byteswap().newbyteorder()

                    break

        else:
            raise RuntimeError(f"No data was found in {fits_file}.")

        images = np.nan_to_num(images)

        if images.ndim == 4 and not self.m_ifs_data:
            raise ValueError('The input data is 4D but ifs_data is set to False. Reading in 4D '
                             'data is only possible by setting the argument to True.')

        if images.ndim < 3 and self.m_ifs_data:
            raise ValueError('It is only possible to read 3D or 4D data when ifs_data is set to '
                             'True.')

        if self.m_overwrite and self.m_image_out_port.tag not in overwrite_tags:
            overwrite_tags.append(self.m_image_out_port.tag)

            if self.m_ifs_data:
                self.m_image_out_port.set_all(images, data_dim=4)
            else:
                self.m_image_out_port.set_all(images, data_dim=3)

            self.m_image_out_port.del_all_attributes()

        else:
            if self.m_ifs_data:
                self.m_image_out_port.append(images, data_dim=4)
            else:
                self.m_image_out_port.append(images, data_dim=3)

        header = hdu_list[0].header

        fits_header = []
        for key in header:
            fits_header.append(f'{key} = {header[key]}')

        hdu_list.close()

        header_out_port = self.add_output_port('fits_header/'+os.path.basename(fits_file))
        header_out_port.set_all(fits_header)

        return header, images.shape

    @typechecked
    def _txt_file_list(self) -> list:
        """
        Internal function to import a list of FITS files from a text file.
        """

        with open(self.m_filenames) as file_obj:
            files = file_obj.readlines()

        # remove newlines
        files = [x.strip() for x in files]

        # remove of empty lines
        files = filter(None, files)

        return list(files)

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Looks for all FITS files in the input directory and imports the
        images into the database. Note that previous database information is overwritten if
        ``overwrite=True``. The filenames are stored as attributes.

        Returns
        -------
        NoneType
            None
        """

        files = []

        if isinstance(self.m_filenames, str):
            files = self._txt_file_list()

            for item in files:
                if not os.path.isfile(item):
                    raise ValueError(f'The file {item} does not exist. Please check that the '
                                     f'path is correct.')

        elif isinstance(self.m_filenames, list):
            files = self.m_filenames

            for item in files:
                if not os.path.isfile(item):
                    raise ValueError(f'The file {item} does not exist. Please check that the '
                                     f'path is correct.')

        elif isinstance(self.m_filenames, type(None)):
            for filename in os.listdir(self.m_input_location):
                if filename.endswith('.fits') and not filename.startswith('._'):
                    files.append(os.path.join(self.m_input_location, filename))

            assert files, 'No FITS files found in %s.' % self.m_input_location

        files.sort()

        overwrite_tags = []
        first_index = 0

        start_time = time.time()
        for i, fits_file in enumerate(files):
            progress(i, len(files), 'Reading FITS files...', start_time)

            header, shape = self.read_single_file(fits_file, overwrite_tags)

            if len(shape) == 2:
                nimages = 1

            elif len(shape) == 3:
                if self.m_ifs_data:
                    nimages = 1
                else:
                    nimages = shape[0]

            elif len(shape) == 4:
                nimages = shape[1]

            else:
                raise ValueError('Data read from FITS file has an invalid shape.')

            set_static_attr(fits_file=fits_file,
                            header=header,
                            config_port=self._m_config_port,
                            image_out_port=self.m_image_out_port,
                            check=self.m_check)

            set_nonstatic_attr(header=header,
                               config_port=self._m_config_port,
                               image_out_port=self.m_image_out_port,
                               check=self.m_check)

            set_extra_attr(fits_file=fits_file,
                           nimages=nimages,
                           config_port=self._m_config_port,
                           image_out_port=self.m_image_out_port,
                           first_index=first_index)

            first_index += nimages

            self.m_image_out_port.flush()

        self.m_image_out_port.close_port()



import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.extract import StarExtractionModule, ExtractBinaryModule
from pynpoint.util.tests import create_config, create_star_data, create_fake_data, remove_test_data


class TestExtract:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_star_data(self.test_dir+'star')
        create_fake_data(self.test_dir+'binary')
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(path=self.test_dir, folders=['star', 'binary'])

    def test_read_data(self) -> None:

        module = FitsReadingModule(name_in='read1',
                                   image_tag='star',
                                   input_dir=self.test_dir+'star',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read1')

        data = self.pipeline.get_data('star')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

        module = FitsReadingModule(name_in='read2',
                                   image_tag='binary',
                                   input_dir=self.test_dir+'binary',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read2')

        data = self.pipeline.get_data('binary')
        assert np.sum(data) == pytest.approx(11.012854046962481, rel=self.limit, abs=0.)
        assert data.shape == (10, 21, 21)

        self.pipeline.set_attribute('binary', 'PARANG', -1.*np.linspace(0., 180., 10), static=False)

    def test_extract_position_none(self) -> None:

        module = StarExtractionModule(name_in='extract1',
                                      image_in_tag='star',
                                      image_out_tag='extract1',
                                      index_out_tag='index',
                                      image_size=0.2,
                                      fwhm_star=0.1,
                                      position=None)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('extract1')

        assert len(warning) == 3

        assert warning[0].message.args[0] == 'Can not store the attribute \'INSTRUMENT\' because ' \
                                             'the dataset \'index\' does not exist.'

        assert warning[1].message.args[0] == 'Can not store the attribute \'PIXSCALE\' because ' \
                                             'the dataset \'index\' does not exist.'

        assert warning[2].message.args[0] == 'Can not store the attribute \'History: ' \
                                             'StarExtractionModule\' because the dataset ' \
                                             '\'index\' does not exist.'

        data = self.pipeline.get_data('extract1')
        assert np.sum(data) == pytest.approx(104.93318507061295, rel=self.limit, abs=0.)
        assert data.shape == (10, 9, 9)

    def test_extract_center_none(self) -> None:

        module = StarExtractionModule(name_in='extract2',
                                      image_in_tag='star',
                                      image_out_tag='extract2',
                                      index_out_tag='index',
                                      image_size=0.2,
                                      fwhm_star=0.1,
                                      position=(None, None, 0.2))

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('extract2')

        assert len(warning) == 3

        assert warning[0].message.args[0] == 'Can not store the attribute \'INSTRUMENT\' because ' \
                                             'the dataset \'index\' does not exist.'

        assert warning[1].message.args[0] == 'Can not store the attribute \'PIXSCALE\' because ' \
                                             'the dataset \'index\' does not exist.'

        assert warning[2].message.args[0] == 'Can not store the attribute \'History: ' \
                                             'StarExtractionModule\' because the dataset ' \
                                             '\'index\' does not exist.'

        data = self.pipeline.get_data('extract2')
        assert np.sum(data) == pytest.approx(104.93318507061295, rel=self.limit, abs=0.)
        assert data.shape == (10, 9, 9)

    def test_extract_position(self) -> None:

        module = StarExtractionModule(name_in='extract7',
                                      image_in_tag='star',
                                      image_out_tag='extract7',
                                      index_out_tag=None,
                                      image_size=0.2,
                                      fwhm_star=0.1,
                                      position=(5, 5, 0.2))

        self.pipeline.add_module(module)
        self.pipeline.run_module('extract7')

        data = self.pipeline.get_data('extract7')
        assert np.sum(data) == pytest.approx(104.93318507061295, rel=self.limit, abs=0.)
        assert data.shape == (10, 9, 9)

    def test_extract_too_large(self) -> None:

        module = StarExtractionModule(name_in='extract3',
                                      image_in_tag='star',
                                      image_out_tag='extract3',
                                      index_out_tag=None,
                                      image_size=0.2,
                                      fwhm_star=0.1,
                                      position=(2, 2, 0.05))

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('extract3')

        assert len(warning) == 10

        assert warning[0].message.args[0] == f'Chosen image size is too large to crop the image ' \
                                             f'around the brightest pixel (image index = 0, ' \
                                             f'pixel [x, y] = [2, 2]). Using the center of ' \
                                             f'the image instead.'

        data = self.pipeline.get_data('extract3')
        assert np.sum(data) == pytest.approx(104.93318507061295, rel=self.limit, abs=0.)
        assert data.shape == (10, 9, 9)

    def test_star_extract_cpu(self) -> None:

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file['config'].attrs['CPU'] = 4

        module = StarExtractionModule(name_in='extract4',
                                      image_in_tag='star',
                                      image_out_tag='extract4',
                                      index_out_tag='index',
                                      image_size=0.2,
                                      fwhm_star=0.1,
                                      position=(2, 2, 0.05))

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('extract4')

        assert len(warning) == 2

        assert warning[0].message.args[0] == 'The \'index_out_port\' can only be used if ' \
                                             'CPU = 1. No data will be stored to this output port.'

        assert warning[1].message.args[0] == 'Chosen image size is too large to crop the image ' \
                                             'around the brightest pixel (image index = 0, ' \
                                             'pixel [x, y] = [2, 2]). Using the center of the ' \
                                             'image instead.'

    def test_extract_binary(self) -> None:

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file['config'].attrs['CPU'] = 1

        module = ExtractBinaryModule(pos_center=(10., 10.),
                                     pos_binary=(10., 16.),
                                     name_in='extract5',
                                     image_in_tag='binary',
                                     image_out_tag='extract5',
                                     image_size=0.15,
                                     search_size=0.07,
                                     filter_size=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('extract5')

        data = self.pipeline.get_data('extract5')
        assert np.sum(data) == pytest.approx(1.3419098759577548, rel=self.limit, abs=0.)
        assert data.shape == (10, 7, 7)

    def test_extract_binary_filter(self) -> None:

        module = ExtractBinaryModule(pos_center=(10., 10.),
                                     pos_binary=(10., 16.),
                                     name_in='extract6',
                                     image_in_tag='binary',
                                     image_out_tag='extract6',
                                     image_size=0.15,
                                     search_size=0.07,
                                     filter_size=0.05)

        self.pipeline.add_module(module)
        self.pipeline.run_module('extract6')

        data = self.pipeline.get_data('extract6')
        assert np.sum(data) == pytest.approx(1.3789593661036972, rel=self.limit, abs=0.)
        assert data.shape == (10, 7, 7)



import os

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.basic import SubtractImagesModule, AddImagesModule, RotateImagesModule, \
                                      RepeatImagesModule
from pynpoint.util.tests import create_config, remove_test_data, create_star_data


class TestBasic:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_star_data(self.test_dir+'data1')
        create_star_data(self.test_dir+'data2')
        create_star_data(self.test_dir+'data3')

        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['data1', 'data2', 'data3'])

    def test_read_data(self) -> None:

        module = FitsReadingModule(name_in='read1',
                                   image_tag='data1',
                                   input_dir=self.test_dir+'data1',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)

        module = FitsReadingModule(name_in='read2',
                                   image_tag='data2',
                                   input_dir=self.test_dir+'data2',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)

        module = FitsReadingModule(name_in='read3',
                                   image_tag='data3',
                                   input_dir=self.test_dir+'data3',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)

        self.pipeline.run_module('read1')
        self.pipeline.run_module('read2')
        self.pipeline.run_module('read3')

        data = self.pipeline.get_data('data1')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

        data = self.pipeline.get_data('data2')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

        data = self.pipeline.get_data('data3')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_subtract_images(self) -> None:

        module = SubtractImagesModule(name_in='subtract',
                                      image_in_tags=('data1', 'data2'),
                                      image_out_tag='subtract',
                                      scaling=1.)

        self.pipeline.add_module(module)
        self.pipeline.run_module('subtract')

        data = self.pipeline.get_data('subtract')
        assert np.sum(data) == pytest.approx(0., rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_add_images(self) -> None:

        module = AddImagesModule(name_in='add',
                                 image_in_tags=('data1', 'data2'),
                                 image_out_tag='add',
                                 scaling=1.)

        self.pipeline.add_module(module)
        self.pipeline.run_module('add')

        data = self.pipeline.get_data('add')
        assert np.sum(data) == pytest.approx(211.08557759610554, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_rotate_images(self) -> None:

        module = RotateImagesModule(name_in='rotate',
                                    image_in_tag='data1',
                                    image_out_tag='rotate',
                                    angle=10.)

        self.pipeline.add_module(module)
        self.pipeline.run_module('rotate')

        data = self.pipeline.get_data('rotate')
        assert np.sum(data) == pytest.approx(105.86657256219851, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_repeat_images(self) -> None:

        module = RepeatImagesModule(name_in='repeat',
                                    image_in_tag='data1',
                                    image_out_tag='repeat',
                                    repeat=2)

        self.pipeline.add_module(module)
        self.pipeline.run_module('repeat')

        data1 = self.pipeline.get_data('data1')
        assert data1.shape == (10, 11, 11)

        data2 = self.pipeline.get_data('repeat')
        assert data2.shape == (20, 11, 11)

        assert data1 == pytest.approx(data2[0:10, ], rel=self.limit, abs=0.)
        assert data1 == pytest.approx(data2[10:20, ], rel=self.limit, abs=0.)



import os

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.readwrite.fitswriting import FitsWritingModule
from pynpoint.util.tests import create_config, create_star_data, remove_test_data


class TestFitsWriting:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_star_data(self.test_dir+'fits_data')
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['fits_data'], files=['test.fits'])

    def test_fits_reading(self) -> None:

        module = FitsReadingModule(name_in='read',
                                   input_dir=self.test_dir+'fits_data',
                                   image_tag='images',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read')

        data = self.pipeline.get_data('images')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_fits_writing(self) -> None:

        module = FitsWritingModule(file_name='test.fits',
                                   name_in='write1',
                                   output_dir=None,
                                   data_tag='images',
                                   data_range=None,
                                   overwrite=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('write1')

    def test_filename_extension(self) -> None:

        with pytest.raises(ValueError) as error:
            FitsWritingModule(file_name='test.dat',
                              name_in='write3',
                              output_dir=None,
                              data_tag='images',
                              data_range=None,
                              overwrite=True,
                              subset_size=None)

        assert str(error.value) == 'Output \'file_name\' requires the FITS extension.'

    def test_data_range(self) -> None:

        module = FitsWritingModule(file_name='test.fits',
                                   name_in='write4',
                                   output_dir=None,
                                   data_tag='images',
                                   data_range=(0, 10),
                                   overwrite=True,
                                   subset_size=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('write4')

    def test_not_overwritten(self) -> None:

        module = FitsWritingModule(file_name='test.fits',
                                   name_in='write5',
                                   output_dir=None,
                                   data_tag='images',
                                   data_range=None,
                                   overwrite=False,
                                   subset_size=None)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('write5')

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'Filename already present. Use overwrite=True ' \
                                             'to overwrite an existing FITS file.'

    def test_subset_size(self) -> None:

        module = FitsWritingModule(file_name='test.fits',
                                   name_in='write6',
                                   output_dir=None,
                                   data_tag='images',
                                   data_range=None,
                                   overwrite=True,
                                   subset_size=10)

        self.pipeline.add_module(module)
        self.pipeline.run_module('write6')

    def test_subset_size_data_range(self) -> None:

        module = FitsWritingModule(file_name='test.fits',
                                   name_in='write7',
                                   output_dir=None,
                                   data_tag='images',
                                   data_range=(8, 18),
                                   overwrite=True,
                                   subset_size=10)

        self.pipeline.add_module(module)
        self.pipeline.run_module('write7')

    def test_attribute_length(self) -> None:

        text = 'long_text_long_text_long_text_long_text_long_text_long_text_long_text_long_text'

        self.pipeline.set_attribute('images', 'short', 'value', static=True)
        self.pipeline.set_attribute('images', 'longer_than_eight1', 'value', static=True)
        self.pipeline.set_attribute('images', 'longer_than_eight2', text, static=True)

        module = FitsWritingModule(file_name='test.fits',
                                   name_in='write8',
                                   output_dir=None,
                                   data_tag='images',
                                   data_range=None,
                                   overwrite=True,
                                   subset_size=None)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('write8')

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'Key \'hierarch longer_than_eight2\' with value ' \
                                             '\'long_text_long_text_long_text_long_text_long_' \
                                             'text_long_text_long_text_long_text\' is too ' \
                                             'long for the FITS format. To avoid an error, ' \
                                             'the value was truncated to \'long_text_long_text' \
                                             '_long_text_long_text_long_tex\'.'



"""
Module for exporting a dataset from the HDF5 database to a FITS file.
"""

import os
import warnings

from typing import Optional, Tuple

from astropy.io import fits
from typeguard import typechecked

import numpy as np

from pynpoint.core.processing import WritingModule
from pynpoint.util.module import memory_frames


class FitsWritingModule(WritingModule):
    """
    Module for writing a dataset from the central HDF5 database to a FITS file. The static
    attributes will be stored as header information. The dataset is selected from the database
    by its tag name. :class:`~pynpoint.readwrite.fitswriting.FitsWritingModule` is a
    :class:`~pynpoint.core.processing.WritingModule` and uses either the default output directory
    of a :class:`~pynpoint.core.pypeline.Pypeline` or a specified location to store the FITS data.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 data_tag: str,
                 file_name: str,
                 output_dir: Optional[str] = None,
                 data_range: Optional[Tuple[int, int]] = None,
                 overwrite: bool = True,
                 subset_size: Optional[int] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        data_tag : str
            Tag of the database entry that has to be exported to a FITS file.
        file_name : str
            Name of the FITS output file. Requires the FITS extension.
        output_dir : str, None
            Output directory where the FITS file will be stored. If no folder is specified the
            Pypeline default is chosen.
        data_range : tuple, None
            A two element tuple which specifies a begin and end frame of the export. This can be
            used to save a subsets of a large dataset. The whole dataset will be exported if set
            to None.
        overwrite : bool
            Overwrite an existing FITS file with an identical filename.
        subset_size : int, None
            Size of the subsets that are created when storing the data. This can be useful if the
            dataset contains a large number of images. An increasing index value is appended to
            the FITS file names. All images are written to a single FITS file if set to None.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in, output_dir=output_dir)

        if not file_name.endswith('.fits'):
            raise ValueError('Output \'file_name\' requires the FITS extension.')

        self.m_file_name = file_name
        self.m_data_port = self.add_input_port(data_tag)
        self.m_range = data_range
        self.m_overwrite = overwrite
        self.m_subset_size = subset_size

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Creates a FITS file and stores the data and the corresponding
        static attributes.

        Returns
        -------
        NoneType
            None
        """

        out_name = os.path.join(self.m_output_location, self.m_file_name)

        print('Writing FITS file...', end='')

        if os.path.isfile(out_name) and not self.m_overwrite:
            warnings.warn('Filename already present. Use overwrite=True to overwrite an existing '
                          'FITS file.')

        else:
            header = fits.Header()
            attributes = self.m_data_port.get_all_static_attributes()

            for attr in attributes:

                if len(attr) > 8:

                    # Check if the header keyword together with its value is
                    # too long for the FITS format. If that is the case, raise
                    # a warning and truncate the value to avoid a ValueError.
                    key = 'hierarch ' + attr
                    value = str(attributes[attr])
                    max_val_len = 75 - len(key)

                    if len(key + value) > 75:
                        warnings.warn(f'Key \'{key}\' with value \'{value}\' is too long for '
                                      f'the FITS format. To avoid an error, the value was '
                                      f'truncated to \'{value[:max_val_len]}\'.')

                    header[key] = value[:max_val_len]

                else:
                    header[attr] = attributes[attr]

            if self.m_subset_size is None:
                if self.m_range is None:
                    frames = [0, self.m_data_port.get_shape()[0]]

                else:
                    frames = [self.m_range[0], self.m_range[1]]

            else:
                if self.m_range is None:
                    nimages = self.m_data_port.get_shape()[0]
                    frames = memory_frames(self.m_subset_size, nimages)

                else:
                    nimages = self.m_range[1] - self.m_range[0]
                    frames = memory_frames(self.m_subset_size, nimages)
                    frames = np.asarray(frames) + self.m_range[0]

            for i, _ in enumerate(frames[:-1]):
                data_select = self.m_data_port[frames[i]:frames[i+1], ]

                if len(frames) == 2:
                    fits.writeto(out_name, data_select, header, overwrite=self.m_overwrite)

                else:
                    filename = f'{out_name[:-5]}{i:03d}.fits'
                    fits.writeto(filename, data_select, header, overwrite=self.m_overwrite)

            print(' [DONE]')

        self.m_data_port.close_port()
